import itertools
import random
import sys

from HTSeq.scripts.utils import invert_strand, UnknownChrom
from HTSeq.scripts.count_features.reads_io_processor import ReadsIO
from HTSeq.scripts.count_features.reads_stats import ReadsStatistics


def count_reads_single_file(
    isam,
    sam_filename,
    features,
    feature_attr,
    order,
    max_buffer_size,
    stranded,
    overlap_mode,
    multimapped_mode,
    secondary_alignment_mode,
    supplementary_alignment_mode,
    feature_type,
    id_attribute,
    additional_attributes,
    quiet,
    minaqual,
    samout_format,
    samout_filename,
):
    """
    The function that does the counting for each input BAM/SAM file.
    Fixme: there are some redundant parameters here.. feature_type, id_attribute, additional_attributes

    Parameters
    ----------
    isam : int
        input files' indexing for the purpose of parallel processing.
        This basically tell you which input file is being processed by this
        instance of function.
    sam_filename : str
        Path to the SAM/BAM file containing the mapped reads.
    features : array
        TODO check the type of this parameter.
        Supplied by HTSeq.make_feature_genomicarrayofsets
    feature_attr : array
        TODO check the type of this parameter.
        Supplied by HTSeq.make_feature_genomicarrayofsets
    order : str
        Can only be either 'pos' or 'name'. Sorting order of <alignment_file>.
    max_buffer_size : int
        The number of reads allowed to stay in memory until mates are found.
        Used when <alignment_file> is paired end sorted by position.
    stranded : str
        Whether the data to be aligned is from a strand-specific assay.
        Option is yes, no, reverse.
        reverse means yes with reversed strand interpretation.
    overlap_mode : str
        Mode to handle reads overlapping more than one feature.
        Choices: union, intersection-strict, intersection-nonempty.
    multimapped_mode : str
        Whether and how to score reads that are not uniquely aligned or
        ambiguously assigned to features.
        Choices: none, all, fraction, random.
    secondary_alignment_mode : str
        Whether to score secondary alignments (0x100 flag).
        Choices: score or ignore.
    supplementary_alignment_mode : str
        Whether to score supplementary alignments (0x800 flag).
        Choices: score or ignore.
    feature_type : str
        Feature type (3rd column in GTF file) to be used, all features of other
        type are ignored (default, suitable for Ensembl, GTF files: exon).
    id_attribute : str
        GTF attribute to be used as feature ID.
        Normally gene_id, suitable for Ensembl GTF files.
    additional_attributes : array
        Additional feature attributes.
        Commonly, gene_name is suitable for Ensembl GTF files.
    quiet : boolean
        Whether to suppress progress report.
    minaqual : int
        Value denoting the MAPQ alignment quality of reads to skip.
    samout_format : str
        Format of the output files denoted by samouts.
        Choices: SAM, BAM, sam, bam.
    samout_filename : str
        The name of SAM/BAM file to write out all SAM alignment records into.
    Returns
    -------
    Dictionary
        TODO update me when done refactoring

    """
    try:
        read_io_obj = ReadsIO(
            sam_filename=sam_filename,
            samout_filename=samout_filename,
            samout_format=samout_format,
            supplementary_alignment_mode=supplementary_alignment_mode,
            secondary_alignment_mode=secondary_alignment_mode,
            order=order,
            max_buffer_size=max_buffer_size,
        )

        # If the BAM header is available, check that at least one of the
        # chromosomes is also found in the GTF/GFF file, otherwise the user
        # is probably doing something wrong (e.g. "chr1" vs "1").
        bam_chroms = read_io_obj.get_chromosome_names_header()
        if bam_chroms is not None:
            bam_chroms = set(bam_chroms)
            feature_chroms = set(features.chrom_vectors.keys())
            if not (bam_chroms & feature_chroms):
                sys.stderr.write(
                    f"The alignment file has no chromosomes in common with the GFF/GTF "
                    "file. This will result in zero feature counts. Please check if the "
                    "references match, e.g. if you are using 'chr1' or '1' as "
                    "chromosome names.\n")

    except:
        sys.stderr.write("Error occurred when reading beginning of SAM/BAM file.\n")
        raise

    try:
        read_stats = ReadsStatistics(
            feature_attr=feature_attr, read_io_object=read_io_obj
        )
    except:
        sys.stderr.write(
            "Error occurred when preparing object to store the reads' assignments\n"
        )
        raise

    # CIGAR match characters (including alignment match, sequence match, and
    # sequence mismatch
    com = ("M", "=", "X")

    try:
        for r in read_io_obj.read_seq:
            read_stats.print_progress()
            read_stats.add_num_reads_processed()

            #  get the interval/read sequence.
            if not read_io_obj.pe_mode:
                skip_read = _assess_non_pe_read(
                    read_sequence=r,
                    read_stats=read_stats,
                    secondary_alignment_mode=secondary_alignment_mode,
                    supplementary_alignment_mode=supplementary_alignment_mode,
                    multimapped_mode=multimapped_mode,
                    minaqual=minaqual,
                )

                if skip_read:
                    continue
                iv_seq = _get_iv_seq_non_pe_read(com, r, stranded)
            else:

                # NOTE: the logic here is a little arbitrary and might benefit
                # from an optional arg. If the reads are paired-end but one of
                # the two is missing, ATM we rely on the other one for info,
                # however the data is technically inconsistent and we might
                # want to let the user choose.
                skip_read = _assess_pe_read(
                    minaqual,
                    multimapped_mode,
                    r,
                    read_stats,
                    secondary_alignment_mode,
                    supplementary_alignment_mode,
                )
                if skip_read:
                    continue

                iv_seq = _get_iv_seq_pe_read(com, r, stranded)

            # this bit updates the counts obtained from aligning reads to feature sets.
            try:
                fs = _align_reads_to_feature_set(features, iv_seq, overlap_mode)

                _update_feature_set_counts(fs, multimapped_mode, r, read_stats)

            except UnknownChrom:
                read_stats.add_empty_read(read_sequence=r)

    except:
        sys.stderr.write(
            "Error occured when processing input (%s):\n"
            % (read_io_obj.read_seq_file.get_line_number_string())
        )
        raise

    if not quiet:
        read_stats.print_progress(force_print=True)

    read_io_obj.close_samoutfile()

    res = read_stats.get_output(isam)
    return res


def _update_feature_set_counts(fs, multimapped_mode, read_sequence, read_stats):
    """
    Distribute the counts among the aligned feature set.

    Parameters
    ----------
    fs : array
        A list of feature set previously aligned to the read
    multimapped_mode : str
        How to handle read mapped to multiple features
    read_sequence : array
        Read sequence
    read_stats : ReadsStatistics object
        For updating bad reads

    """
    if fs is None or len(fs) == 0:
        read_stats.add_empty_read(read_sequence=read_sequence)
    elif len(fs) > 1:
        read_stats.add_ambiguous_read(
            read_sequence=read_sequence,
            assignment="__ambiguous[" + "+".join(sorted(fs)) + "]",
        )
    else:
        read_stats.add_good_read_assignment(
            read_sequence=read_sequence, assignment=list(fs)[0]
        )
    if fs is not None and len(fs) > 0:
        fs = list(fs)
        if multimapped_mode == "none":
            if len(fs) == 1:
                read_stats.add_to_count(feature=fs[0])
        elif multimapped_mode == "all":
            for fsi in fs:
                read_stats.add_to_count(feature=fsi)
        elif multimapped_mode == "fraction":
            val = 1.0 / len(fs)
            for fsi in fs:
                read_stats.add_to_count(feature=fsi, value=val)
        elif multimapped_mode == "random":
            fsi = random.choice(fs)
            read_stats.add_to_count(feature=fsi)
        else:
            sys.exit("Illegal multimap mode.")


def _align_reads_to_feature_set(features, iv_seq, overlap_mode):
    """
    Align reads to feature set.

    Parameters
    ----------
    features : array
        A set of features to align the reads to
        TODO not sure the type yet.
    iv_seq : array
        TODO not sure the type yet.
        Read (or interval?) sequence
    overlap_mode : str
        How to select the features for read that not 100% aligned to a feature.

    Returns
    -------
    fs : array
        A set of features to align the reads to
        TODO not sure the type yet.

    """
    if overlap_mode == "union":
        fs = set()
        for iv in iv_seq:
            if iv.chrom not in features.chrom_vectors:
                raise UnknownChrom
            for iv2, fs2 in features[iv].steps():
                fs = fs.union(fs2)
    elif overlap_mode in ("intersection-strict", "intersection-nonempty"):
        fs = None
        for iv in iv_seq:
            if iv.chrom not in features.chrom_vectors:
                raise UnknownChrom
            for iv2, fs2 in features[iv].steps():
                if (len(fs2) > 0) or (overlap_mode == "intersection-strict"):
                    if fs is None:
                        fs = fs2.copy()
                    else:
                        fs = fs.intersection(fs2)
    else:
        sys.exit("Illegal overlap mode.")
    return fs


def _get_iv_seq_pe_read(com, r, stranded):
    """
    Function to break down the read sequence into intervals which will
    subsequently be processed.

    Parameters
    ----------
    com : array
        CIGAR match characters (including alignment match, sequence match, and
        sequence mismatch
    r :
        todo update type
        Read sequence
    stranded : str
        Whether the data to be aligned is from a strand-specific assay.
        Option is yes, no, reverse.
        reverse means yes with reversed strand interpretation.

    Returns
    -------
    iv_seq :
        todo update type

    """
    if r[0] is not None and r[0].aligned:
        iv_seq = _get_iv_seq_pe_read_first(com, r[0], stranded)
    else:
        iv_seq = tuple()
    if r[1] is not None and r[1].aligned:
        iv_seq = _get_iv_seq_pe_read_second(com, iv_seq, r[1], stranded)
    return iv_seq


def _assess_pe_read(
    minaqual,
    multimapped_mode,
    read_sequence,
    read_stats,
    secondary_alignment_mode,
    supplementary_alignment_mode,
):
    """
    Function to check the read for paired end.

    Parameters
    ----------
    minaqual : int
        Value denoting the MAPQ alignment quality of reads to skip.
    multimapped_mode : str
        Whether and how to score reads that are not uniquely aligned or
        ambiguously assigned to features.
        Choices: none, all, fraction, random.
    read_sequence :
        todo update type
    read_stats : ReadsStatistics object
        Object which stores a bunch of statistics about the read sequences.
    secondary_alignment_mode : str
        Whether to score secondary alignments (0x100 flag).
        Choices: score or ignore.
    supplementary_alignment_mode : str
        Whether to score supplementary alignments (0x800 flag).
        Choices: score or ignore.

    Returns
    -------

    """
    # NOTE: Sometimes read1 is None or not aligned but read2 is fine, in that
    # case we should not exclude the entire pair but rather use the interval
    # of the second read
    read1_miss = (read_sequence[0] is None) or (not read_sequence[0].aligned)
    read2_miss = (read_sequence[1] is None) or (not read_sequence[1].aligned)
    if read1_miss and read2_miss:
        read_stats.add_not_aligned_read(read_sequence=read_sequence)
        return True

    if secondary_alignment_mode == "ignore":
        if (read_sequence[0] is not None) and read_sequence[0].not_primary_alignment:
            return True
        elif (read_sequence[1] is not None) and read_sequence[1].not_primary_alignment:
            return True
    if supplementary_alignment_mode == "ignore":
        if (read_sequence[0] is not None) and read_sequence[0].supplementary:
            return True
        elif (read_sequence[1] is not None) and read_sequence[1].supplementary:
            return True
    try:
        if (
            read_sequence[0] is not None and read_sequence[0].optional_field("NH") > 1
        ) or (
            read_sequence[1] is not None and read_sequence[1].optional_field("NH") > 1
        ):
            read_stats.add_not_unique_read(read_sequence=read_sequence)
            if multimapped_mode == "none":
                return True
    except KeyError:
        pass
    if (read_sequence[0] and read_sequence[0].aQual < minaqual) or (
        read_sequence[1] and read_sequence[1].aQual < minaqual
    ):
        read_stats.add_low_quality_read(read_sequence=read_sequence)
        return True
    return False


# Get GenomicInterval for each read, whether single-end or paired-end
def _get_iv_seq_non_pe_read(com, r, stranded):
    if stranded != "reverse":
        iv_seq = (co.ref_iv for co in r.cigar if co.type in com and co.size > 0)
    else:
        iv_seq = (
            invert_strand(co.ref_iv)
            for co in r.cigar
            if (co.type in com and co.size > 0)
        )
    return iv_seq


def _get_iv_seq_pe_read_first(com, read, stranded):
    if stranded != "reverse":
        iv_seq = (co.ref_iv for co in read.cigar if co.type in com and co.size > 0)
    else:
        iv_seq = (
            invert_strand(co.ref_iv)
            for co in read.cigar
            if co.type in com and co.size > 0
        )
    return iv_seq


def _get_iv_seq_pe_read_second(com, iv_seq, read, stranded):
    if stranded != "reverse":
        iv_seq = itertools.chain(
            iv_seq,
            (
                invert_strand(co.ref_iv)
                for co in read.cigar
                if co.type in com and co.size > 0
            ),
        )
    else:
        iv_seq = itertools.chain(
            iv_seq, (co.ref_iv for co in read.cigar if co.type in com and co.size > 0)
        )
    return iv_seq


def _assess_non_pe_read(
    read_sequence,
    read_stats,
    secondary_alignment_mode,
    supplementary_alignment_mode,
    multimapped_mode,
    minaqual,
):
    if not read_sequence.aligned:
        read_stats.add_not_aligned_read(read_sequence=read_sequence)
        return True
    if (secondary_alignment_mode == "ignore") and read_sequence.not_primary_alignment:
        return True
    if (supplementary_alignment_mode == "ignore") and read_sequence.supplementary:
        return True
    try:
        if read_sequence.optional_field("NH") > 1:
            read_stats.add_not_unique_read(read_sequence=read_sequence)
            if multimapped_mode == "none":
                return True
    except KeyError:
        pass
    if read_sequence.aQual < minaqual:
        read_stats.add_low_quality_read(read_sequence=read_sequence)
        return True

    return False



import sys
import os
import math
import re
import csv
import gzip
import itertools
import collections
try:
    import cStringIO
except:
    pass
import warnings
import numpy
cimport numpy

from HTSeq import StepVector
from HTSeq.StretchVector import StretchVector
from HTSeq import _HTSeq_internal


###########################
# GenomicInterval
###########################

cdef str strand_plus = intern("+")
cdef str strand_minus = intern("-")
cdef str strand_nostrand = intern(".")


cdef class GenomicInterval:
    """A range of consecutive positions on a reference genome.

        Properties:

        chrom: The name of a sequence (i.e., chromosome, contig, etc.).
        start: The start of the interval. Even on the reverse strand,
          this is always the smaller of the two values 'start' and 'end'.
          Note that all positions should be given as 0-based value!
        end: The end of the interval. Following Python convention for
          ranges, this in one more than the coordinate of the last base
          that is considered part of the sequence.
        strand: The strand, as a single character, '+' or '-'. '.' indicates
          that the strand is irrelavant. (Alternatively, pass a Strand object.)
        length: The length of the interval, i.e., end - start
        start_d: The "directional start" position. This is the position of the
          first base of the interval, taking the strand into account. Hence,
          this is the same as 'start' except when strand == '-', in which
          case it is end-1.
        end_d: The "directional end": Usually, the same as 'end', but for
          strand=='-1', it is start-1.
    """

    def __init__(GenomicInterval self, str chrom, long start, long end,
                 str strand=strand_nostrand):
        """See the class docstring for the meaning of the slots. Note that
        there is also a factory function, 'from_directional', to be used if
        you wish to specify start_d and length.
        """
        self.chrom = intern(chrom)
        self.start = start
        self.end = end
        self.strand = strand
        if self.start > self.end:
            raise ValueError("start is larger than end")

    property strand:
        def __set__(self, strand):
            strand = intern(strand)
            if not(strand is strand_plus or strand is strand_minus or
                    strand is strand_nostrand):
                raise ValueError("Strand must be'+', '-', or '.'.")
            self._strand = strand

        def __get__(self):
            return self._strand

    def __reduce__(GenomicInterval self):
        return GenomicInterval, (self.chrom, self.start, self.end,
                                 self.strand)

    def __copy__(self):
        constr, args = self.__reduce__()
        return constr(*args)

    def __repr__(GenomicInterval self):
        return "<%s object '%s', [%d,%s), strand '%s'>" % \
            (self.__class__.__name__, self.chrom, self.start,
             str(self.end) if self.end != sys.maxsize else "Inf", self.strand)

    def __str__(GenomicInterval self):
        return "%s:[%d,%s)/%s" % \
            (self.chrom, self.start, str(self.end)
             if self.end != sys.maxsize else "Inf", self.strand)

    property length:

        """The length is calculated as end - start. If you set the length,
        'start_d' will be preserved, i.e., 'end' is changed, unless the strand
        is '-', in which case 'start' is changed."""

        def __get__(GenomicInterval self):
            return self.end - self.start

        def __set__(GenomicInterval self, long newLength):
            if self._strand is not strand_minus:
                self.end = self.start + newLength
            else:
                self.start = self.end - newLength

    property start_d:
        """See the class docstring for the meaning of the 'directional start'.
        Note that if you set 'start_d', both the start and the end are changed,
        such the interval gets the requested new directional start and its
        length stays unchanged."""

        def __get__(GenomicInterval self):
            if self._strand is not strand_minus:
                return self.start
            else:
                return self.end - 1

        def __set__(GenomicInterval self, long newStartd):
            if self._strand is not strand_minus:
                self.end = newStartd + self.length
                self.start = newStartd
            else:
                self.start = newStartd + 1 - self.length
                self.end = newStartd + 1

    property end_d:

        def __get__(GenomicInterval self):
            if self._strand is not strand_minus:
                return self.end
            else:
                return self.start - 1

    property start_as_pos:
        def __get__(GenomicInterval self):
            return GenomicPosition(self.chrom, self.start, self. strand)

    property end_as_pos:
        def __get__(GenomicInterval self):
            return GenomicPosition(self.chrom, self.end, self. strand)

    property start_d_as_pos:
        def __get__(GenomicInterval self):
            return GenomicPosition(self.chrom, self.start_d, self. strand)

    property end_d_as_pos:
        def __get__(GenomicInterval self):
            return GenomicPosition(self.chrom, self.end_d, self. strand)

    def __richcmp__(GenomicInterval self, GenomicInterval other, int op):
        if op == 2:  # ==
            if other == None:
                return False
            return self._strand is other._strand and \
                self.start == other.start and self.end == other.end
        elif op == 3:  # !=
            return not (self == other)
        else:
            raise NotImplementedError

    def __hash__(GenomicInterval self):
        return hash((self.chrom, self.start, self.end, self.strand))

    cpdef is_contained_in(GenomicInterval self, GenomicInterval iv):
        """Returns a boolean value indicating whether the 'self' interval
        is fully within the 'iv' interval.

        This is deemed the case if
          - both are on the same chromosome, and
          - both are on the same strand, or at least one of them is
             not stranded (i.e., has strand == '.'), and
          - self.start >= iv.start, and
          - self.end <= iv.end
        """
        if iv == None:
            return False
        if self.chrom != iv.chrom:
            return False
        if self._strand is not strand_nostrand and iv.strand is not strand_nostrand and \
                self.strand is not iv._strand:
            return False
        if self.start < iv.start or self.end > iv.end:
            return False
        return True

    cpdef contains(GenomicInterval self, GenomicInterval iv):
        """Returns a boolean value indicating whether the 'self' interval
        fully contains the 'iv' interval.

        See 'is_contained_in' for the exact criteria.
        """
        if iv == None:
            return False
        return iv.is_contained_in(self)

    cpdef overlaps(GenomicInterval self, GenomicInterval iv):
        """Returns a boolean value indicating whether the 'self' interval
        overlaps the 'iv' interval.

        This is deemed the case if
          - both are on the same chromosome, and
          - both are on the same strand, or at least one of them is
             not stranded (i.e., has strand == '.'), and
          - the actual intervals overlap
        """
        if iv == None:
            return False
        if self.chrom != iv.chrom:
            return False
        if self.strand is not strand_nostrand and iv.strand is not strand_nostrand and \
                self.strand is not iv.strand:
            return False
        if self.start <= iv.start:
            return self.end > iv.start
        else:
            return iv.end > self.start

    def range(GenomicInterval self, long int step=1):
        """Generate an iterator over the GenomicPositions covered by the interval,
        running from start to end.
        """
        return _HTSeq_internal.GenomicInterval_range(self, step)

    def range_d(GenomicInterval self, long int step=1):
        """Generate an iterator over the GenomicPositions covered by the interval.
        running from start_d to end_d.
        """
        return _HTSeq_internal.GenomicInterval_ranged(self, step)

    cpdef extend_to_include(GenomicInterval self, GenomicInterval iv):
        """Extend the interval such that it includes iv."""
        if iv is None:
            raise TypeError("Cannot extend an interval to include None.")
        if self.chrom != iv.chrom:
            raise ValueError("Cannot extend an interval to include an interval on another chromosome.")
        if self.strand is not strand_nostrand and iv.strand is not strand_nostrand and \
                self.strand is not iv.strand:
            raise ValueError("Cannot extend an interval to include an interval on another strand.")
        self.start = min(self.start, iv.start)
        self.end = max(self.end, iv.end)

    def copy(self):
        return GenomicInterval(self.chrom, self.start, self.end, self.strand)


def GenomicInterval_from_directional(str chrom, long int start_d, long int length, str strand="."):
    strand = intern(strand)
    if strand != strand_minus:
        return GenomicInterval(chrom, start_d, start_d + length, strand)
    else:
        return GenomicInterval(chrom, start_d - length + 1, start_d + 1, strand)


cdef class GenomicPosition(GenomicInterval):
    """Position of a nucleotide or base pair on a reference genome.

    Properties:
       chrom: The name of a sequence (i.e., chromosome, contig, etc.).
       pos: The position on the sequence specified by seqname.
          The position should always be given as 0-based value!
       strand: The strand, as a single character, '+' or '-'. '.' indicates
          that the strand is irrelavant.

    The GenomicPosition class is derived from GenomicInterval. Hence,
    a GenomicPosition is always a GenomicInterval of length 1. Do not tinker
    with the exposed GenomeInterval properties.
    """

    def __init__(self, str chrom, long int pos, str strand='.'):
        GenomicInterval.__init__(self, chrom, pos, pos + 1, strand)

    property pos:
        """Alias for 'start_d'."""

        def __get__(self):
            return self.start_d

        def __set__(self, long newValue):
            self.start_d = newValue

    property end:

        def __get__(self):
            return self.start + 1

    property length:

        def __get__(self):
            return 1

    def __repr__(self):
        return "<%s object '%s':%d, strand '%s'>" % \
            (self.__class__.__name__, self.chrom, self.pos, self.strand)

    def __str__(self):
        return "%s:%d/%s" % (self.chrom, self.pos, self.strand)

    def __reduce__(GenomicPosition self):
        return GenomicPosition, (self.chrom, self.pos, self.strand)

    def copy(self):
        return GenomicPosition(self.chrom, self.pos, self.strand)


cdef class ChromVector(object):
    """Counting vector covering a chromosome.

    This class supports three types of storage:
      1. 'ndarray': Use a dense 1D numpy array
      2. 'memmap': Use numpy memory maps on disk for large arrays
      3. 'step': Use a StepVector
    """

    cdef public object array
    cdef public GenomicInterval iv
    cdef public int offset
    cdef public bint is_vector_of_sets
    cdef public str _storage
    cdef public str typecode
    cdef public str memmap_dir

    @classmethod
    def create(cls, GenomicInterval iv, str typecode, str storage, str memmap_dir=""):
        """Create ChromVector from GenomicInterval

        Args:
            iv (GenomicInterval): A GenomicInterval describing the chromosome
              vector.
            typecode ('d', 'i', 'l', 'b', or 'O'): What kind of data will be
              stored inside this chromosome vector. 'd' for double, 'i' for int,
              'l' for long int, 'b' for boolean, 'O' for arbitrary objects
              (e.g. sets).
            storage ('step', 'ndarray', or 'memmap'): What kind of storage to
              use. 'ndarray' is appropriate for short chromosomes and stores
              each position in the genome into memory. 'memmap' stores all
              positions, but maps the memory onto disk for larger chromosomes.
              'step' is a sparse representation similar to CSR matrices whereby
              only the boundaries between genomic stretches with differing
              data content are stored - see HTSeq.StepVector.
            memmap_dir (str): If using 'memmap' storage, what folder to store
              the memory maps. These can get quite big.

        Returns:
            An instance of ChromVector with the requested options.

        """
        ncv = cls()
        ncv.iv = iv

        if storage == "ndarray":
            if typecode != 'O':
                ncv.array = numpy.zeros(shape=(iv.length,), dtype=typecode)
            else:
                ncv.array = numpy.empty(shape=(iv.length,), dtype=typecode)
                ncv.array[:] = None

        elif storage == "memmap":
            ncv.array = numpy.memmap(
                shape=(iv.length,),
                dtype=typecode,
                filename=os.path.join(
                    memmap_dir,
                    iv.chrom + iv.strand + str(iv.start) + '_' \
                        + str(iv.length) + ".nmm"),
                mode='w+')

        elif storage == "step":
            ncv.array = StepVector.StepVector.create(
                typecode=typecode,
            )
        elif storage == "stretch":
            ncv.array = StretchVector(
                    typecode=typecode,
            )

        else:
            raise ValueError("Illegal storage mode.")

        ncv._storage = storage
        ncv.typecode = typecode
        # NOTE: As long as autochromosomes in GenomicArray are infinite length
        # this has pretty limited use, but that might change
        ncv.offset = iv.start
        ncv.is_vector_of_sets = False
        ncv.memmap_dir = memmap_dir
        return ncv

    @classmethod
    def _create_view(cls, ChromVector vec, GenomicInterval iv):
        if iv.length == 0:
            raise IndexError("Cannot subset to zero-length interval.")
        v = cls()
        v.iv = iv
        v.array = vec.array
        v.offset = vec.offset
        v.is_vector_of_sets = vec.is_vector_of_sets
        v._storage = vec._storage
        return v

    def extend_to_include(self, iv):
        if iv.strand != self.iv.strand:
            raise ValueError(
                'The new interval must match the current strandedness',
            )

        # Step 1: extend the interval
        length = self.iv.length
        startdiff = max(self.iv.start - iv.start, 0)
        self.iv.extend_to_include(iv)
        self.offset = self.iv.start

        # Step 2: extend the array if needed, and shift-copy the old values
        if self._storage == 'ndarray':
            if self.typecode != 'O':
                array = numpy.zeros(shape=(self.iv.length,), dtype=self.typecode)
            else:
                array = numpy.empty(shape=(self.iv.length,), dtype=self.typecode)
                array[:] = None
            array[startdiff: startdiff + length] = self.array[:]
        elif self._storage == 'memmap':
            array = numpy.memmap(
                shape=(self.iv.length,), dtype=self.typecode,
                filename=os.path.join(
                    self.memmap_dir,
                    self.iv.chrom + self.iv.strand + str(self.iv.start) + '_' \
                        + str(self.iv.length) + ".nmm"),
                mode='w+',
            )
            array[startdiff: startdiff + length] = self.array[:]
        else:
            # The StepVector is created in ChromVector.create without explicit
            # boundaries, so it's already bound by 0, +inf. So we do not need
            # to extend it here, but rather just set the slice to the right
            # value
            array = self.array
        self.array = array

    def __getitem__(self, index):
        """Index or slice the chromosome.

        The index can be a few things:
        - an integer: get the value of the vector at that chromosome coordinate
        - a 1-step slice e.g "4:7": get a view of the chromosome region
          between those coordinates. The array data are not copied.
        - a GenomicInterval: similar to slices, with the additional choice of
          strandedness. If this argument is stranded but the chromosome itself
          is not stranded, a nonstranded view of the chromosome region is
          returned.

        """
        cdef slice index_slice
        cdef long int index_int
        cdef long int start, stop
        cdef GenomicInterval iv

        if isinstance(index, int):
            index_int = index
            if index_int < self.iv.start or index_int >= self.iv.end:
                raise IndexError
            return self.array[index_int - self.offset]

        elif isinstance(index, slice):
            index_slice = index
            if index_slice.start is None:
                start = self.iv.start
            else:
                start = index_slice.start
                if start < self.iv.start:
                    raise IndexError("start too small")

            if index_slice.stop is None:
                stop = self.iv.end
            else:
                stop = index_slice.stop
                if stop > self.iv.end:
                    raise IndexError("stop too large")

            iv = GenomicInterval(self.iv.chrom, start, stop, self.iv.strand)

            if not self.iv.contains(iv):
                raise IndexError
            return ChromVector._create_view(self, iv)

        elif isinstance(index, GenomicInterval):
            if not self.iv.contains(index):
                raise IndexError

            if self.iv.strand is strand_nostrand and \
                    index.strand is not strand_nostrand:
                iv = index.copy()   # Is this correct now?
                iv.strand = strand_nostrand
            else:
                iv = index

            return ChromVector._create_view(self, iv)

        else:
            raise TypeError("Illegal index type")

    def __setitem__(self, index, value):
        cdef slice index_slice
        cdef long int start, stop

        if isinstance(value, ChromVector):
            if self.array is value.array and value.iv.start == index.start and \
                    value.iv.end == index.stop and (index.step is None or index.step == 1):
                return
            else:
                raise NotImplementedError(
                    "Required assignment signature not yet implemented.")

        if isinstance(index, int):
            self.array[index - self.iv.start] = value

        elif isinstance(index, slice):
            index_slice = index
            if index_slice.start is not None:
                start = index_slice.start
                if start < self.iv.start:
                    raise IndexError("start too small")
            else:
                start = self.iv.start
            if index_slice.stop is not None:
                stop = index_slice.stop
                if stop > self.iv.end:
                    raise IndexError("stop too large")
            else:
                stop = self.iv.end
            if start > stop:
                raise IndexError("Start of interval is after its end.")
            if start == stop:
                raise IndexError("Cannot assign to zero-length interval.")
            self.array[start - self.offset: stop -
                       self.iv.start: index.step] = value

        elif isinstance(index, GenomicInterval):
            if index.chrom != self.iv.chrom:
                raise KeyError("Chromosome name mismatch.")
            if self.iv.strand is not strand_nostrand and \
                    self.iv.strand is not self.index.strand:
                raise KeyError("Strand mismatch.")
            self.array[index.iv.start - self.iv.start,
                       index.iv.end - self.iv.start] = value
        else:
            raise TypeError("Illegal index type")

    def __iadd__(self, value):
        if not self.is_vector_of_sets:
            self.array[self.iv.start - self.offset: self.iv.end -
                       self.offset].__iadd__(value)
        else:
            def addval(x):
                y = x.copy()
                y.add(value)
                return y

            self.apply(addval)
        return self

    def __iter__(self):
        return self.values()

    def values(self):
        return iter(self.array[self.iv.start - self.offset: self.iv.end - self.offset])

    def steps(self):
        return _HTSeq_internal.ChromVector_steps(self)

    def apply(self, fun):
        for iv, value in self.steps():
            self.array[iv.start - self.offset: iv.end -
                       self.offset] = fun(value)

    def __repr__(self):
        return "<%s object, %s, %s>" % (self.__class__.__name__, str(self.iv), self._storage)

    def __reduce__(self):
        assert self.__class__ is ChromVector
        return(_ChromVector_unpickle,
               (self.array, self.iv, self.offset, self.is_vector_of_sets, self._storage))


def _ChromVector_unpickle(array, iv, offset, is_vector_of_sets, _storage):
    cv = ChromVector()
    cv.array = array
    cv.iv = iv
    cv.offset = offset
    cv.is_vector_of_sets = is_vector_of_sets
    cv._storage = _storage
    return cv


cdef class GenomicArray(object):
    """Coverage vector including multiple chromosomes.

    This object is basically a collection of ChromVector, with the same options
    for storage:
      1. 'ndarray': Use a dense 1D numpy array
      2. 'memmap': Use numpy memory maps on disk for large arrays
      3. 'step': Use a StepVector

    The class also supports autodiscovery of chromosomes if the 'step' storage
    method is used. In that case, chromosomes of at least sufficient size will
    be created whenever the data pushed into the GenomicArray refers to them.
    For instance, if you are computing plain read coverage along chromosomes,
    each read will inform the GenomicArray as of its chromosome and position:
    the GenomicArray will then create an appropriate ChromVector object of at
    least that size.
    """

    cdef public dict chrom_vectors
    cdef readonly bint stranded
    cdef readonly str typecode
    cdef public str auto_add_chroms
    cdef readonly str storage
    cdef readonly str memmap_dir
    cdef public str header

    def __init__(self, object chroms, bint stranded=True, str typecode='d',
                 str storage='step', str memmap_dir="", str header=""):
        '''GenomicArray(chroms, stranded=True, typecode="d", storage="step", memmap_dir="")

        Initialize GenomicArray

        Args:
            chroms (str, list, or dict): Chromosomes in the GenomicArray. If
              'auto', make infinitely long chromosomes upon the first get or
              set. If 'auto-write', never make new chromosomes upon a get but
              make large enough chromosomes upon the first set.
              If a list, make infinitely long chromosomes with those names.
              If a dict, keys are chromosome names and
              values are their lengths in base pairs. The first two options are
              only available for the 'step' storage (see below).
            stranded (bool): whether the array stores strandedness information.
            typecode ('d', 'i', 'l', 'b', 'O'): what kind of data the array
              will contain. 'd' for double, 'i' for int, 'l' for long int, 'b'
              for boolean, 'O' for arbitrary objects (e.g. sets).
            storage ('step', 'stretch', 'ndarray', or 'memmap'): What kind of
              storage to use. 'ndarray' is appropriate for short chromosomes
              and stores each position in the genome into memory. 'memmap'
              stores all positions, but maps the memory onto disk for larger
              chromosomes. 'step' is a sparse representation similar to CSR
              matrices whereby only the boundaries between genomic stretches
              with differing data content are stored - see HTSeq.StepVector.
              'stretch' is a sparse representation with rare, dense 'islands'
              of data in a sea of missing data along chromosomes.
            memmap_dir (str): If using 'memmap' storage, what folder to store
              the memory maps. These can get quite big.
            header (str): A header with metadata (e.g. when parsing a BedGraph
              file, having the header helps writing it out with all browser
              options retained).

        Returns:
            An instance of GenomicArray with the requested options.
        '''


        self.auto_add_chroms = chroms if chroms in ('auto', 'auto-write') else ''
        self.chrom_vectors = {}
        self.stranded = stranded
        self.typecode = typecode
        self.header = header

        if self.auto_add_chroms:
            chroms = []
            if storage not in ('step', 'stretch'):
                raise TypeError("Automatic adding of chromosomes can " + \
                    " only be used with storage type 'step' or 'stretch'.")

        elif isinstance(chroms, list):
            if storage not in ('step', 'stretch'):
                raise TypeError("Indefinite-length chromosomes can " + \
                    " only be used with storage type 'step' or 'stretch'.")
            chroms = dict([(c, sys.maxsize) for c in chroms])

        elif not isinstance(chroms, dict):
            raise TypeError("'chroms' must be a list or a dict or 'auto'.")

        self.storage = storage
        self.memmap_dir = memmap_dir

        for chrom in chroms:
            self.add_chrom(chrom, chroms[chrom])

    def __getitem__(self, index):
        if isinstance(index, GenomicInterval):
            if self.stranded and index.strand not in (strand_plus, strand_minus):
                raise KeyError(
                    "Non-stranded index used for stranded GenomicArray.")

            # Auto-add chromosome: always infinite size
            if (self.auto_add_chroms == 'auto') and index.chrom not in self.chrom_vectors:
                self.add_chrom(index.chrom)

            if isinstance(index, GenomicPosition):
                if self.stranded:
                    return self.chrom_vectors[index.chrom][index.strand][index.pos]
                else:
                    return self.chrom_vectors[index.chrom][strand_nostrand][index.pos]
            else:
                if self.stranded:
                    return self.chrom_vectors[index.chrom][index.strand][index.start: index.end]
                else:
                    return self.chrom_vectors[index.chrom][strand_nostrand][index.start: index.end]
        else:
            return self.chrom_vectors[index]

    def __setitem__(self, index, value):
        cdef GenomicInterval index2

        if isinstance(value, ChromVector):
            if not isinstance(index, GenomicInterval):
                raise NotImplementedError(
                    "Required assignment signature not yet implemented.")
            index2 = index.copy()
            if not self.stranded:
                index2.strand = strand_nostrand
            if self.chrom_vectors[index2.chrom][index2.strand].array is value.array and index2 == value.iv:
                return
            raise NotImplementedError(
                    "Required assignment signature not yet implemented.")

        if isinstance(index, GenomicInterval):
            if self.stranded and index.strand not in (strand_plus, strand_minus):
                raise KeyError(
                    "Non-stranded index used for stranded GenomicArray.")

            # Auto-add chromosome: always infinite size
            # Before changing this, make sure __setitem__ and __getitem__ are
            # consistent
            if self.auto_add_chroms in ('auto', 'auto-write'):
                # Add a new chromosome
                if index.chrom not in self.chrom_vectors:
                    if self.auto_add_chroms == 'auto':
                        self.add_chrom(index.chrom)
                    else:
                        self.add_chrom(
                                index.chrom,
                                length=index.end - index.start,
                                start_index=index.start,
                        )
                # Extend a known chromosome
                else:
                    if self.stranded:
                        self.chrom_vectors[index.chrom][index.strand].extend_to_include(
                            index,
                        )
                    else:
                        self.chrom_vectors[index.chrom][strand_nostrand].extend_to_include(
                            index,
                        )

            if self.stranded:
                self.chrom_vectors[index.chrom][index.strand][
                    index.start: index.end] = value
            else:
                self.chrom_vectors[index.chrom][strand_nostrand][
                    index.start: index.end] = value
        else:
            raise TypeError("Illegal index type.")

    def add_chrom(self, chrom, length=sys.maxsize, start_index=0):
        cdef GenomicInterval iv
        if length == sys.maxsize:
            iv = GenomicInterval(chrom, start_index, sys.maxsize, ".")
        else:
            iv = GenomicInterval(chrom, start_index, start_index + length, ".")
        if self.stranded:
            self.chrom_vectors[chrom] = {}
            iv.strand = "+"
            self.chrom_vectors[chrom][strand_plus] = \
                ChromVector.create(iv, self.typecode,
                                   self.storage, self.memmap_dir)
            iv = iv.copy()
            iv.strand = "-"
            self.chrom_vectors[chrom][strand_minus] = \
                ChromVector.create(iv, self.typecode,
                                   self.storage, self.memmap_dir)
        else:
            self.chrom_vectors[chrom] = {
                strand_nostrand:  ChromVector.create(iv, self.typecode,
                                                     self.storage,
                                                     self.memmap_dir)}

    def __reduce__(self):
        return (_GenomicArray_unpickle, (self.stranded, self.typecode, self.chrom_vectors))

    def __contains__(self, iv):
        '''Check if the GenomicArray contains a certain interval

        TODO: this is not implemented yet and will throw NotImplementedError.
        '''
        raise NotImplementedError

    def write_bedgraph_file(
            self,
            file_or_filename,
            strand=".",
            track_options="",
            separator='\t',
            ):
        '''Write GenomicArray to BedGraph file

        BedGraph files are used to visualize genomic "tracks", notably in
        UCSC's genomic viewer. This function stores the GenomicArray into such
        a file for further use.

        Args:
            file_or_filename (str, path, or open file handle): Where to store
              the BedGraph data.
            strand ("+", "-", or "."): Which strand to store the array onto.
            track_options (str): A string pre-formatted to describe the track
              options as they appear on the first line of the BedGraph file,
              after "track type=bedGraph".
            separator (str): the pattern that separates the columns.

        The BedGraph file format is described here:

            http://genome.ucsc.edu/goldenPath/help/bedgraph.html
        '''
        sep = separator

        if (not self.stranded) and strand != ".":
            raise ValueError("Strand specified in unstranded GenomicArray.")
        if self.stranded and strand not in (strand_plus, strand_minus):
            raise ValueError("Strand must be specified for stranded GenomicArray.")
        if hasattr(file_or_filename, "write"):
            f = file_or_filename
        else:
            f = open(file_or_filename, "w")

        try:
            if self.header:
                f.write(self.header)
                if not self.header.endswith('\n'):
                    f.write('\n')
            if track_options == "":
                f.write("track type=bedGraph\n")
            else:
                f.write("track type=bedGraph %s\n" % track_options)
            for chrom in self.chrom_vectors:
                for iv, value in self.chrom_vectors[chrom][strand].steps():
                    if iv.start == -sys.maxsize - 1 or iv.end == sys.maxsize:
                        continue
                    f.write(
                        sep.join(
                            (iv.chrom, str(iv.start), str(iv.end), str(value)),
                            )+'\n',
                        )
        finally:
            # Close the file only if we were the ones to open it
            if not hasattr(file_or_filename, "write"):
                f.close()

    @classmethod
    def from_bedgraph_file(cls, file_or_filename, strand=".", typecode="d"):
        '''Create GenomicArray from BedGraph file

        See GenomicArray.write_bedgraph_file for details on the file format.

        Args:
            file_or_filename (str, path, or open file handle): Where to load
              the BedGraph data from.
            strand ("+", "-", or "."): strandedness of the returned array.
            typecode ("d", "i", or "l"): Type of data in the file.
              "d" means floating point (double), "i" is integer, "l" is long
              integer.

        Returns:
            A GenomicArray instance with the data.
        '''
        if hasattr(file_or_filename, "read"):
            f = file_or_filename
        else:
            f = open(file_or_filename, "r")

        try:
            # Find beginning of actual contents
            header = []
            for line in f:
                if line.startswith('track type=bedGraph'):
                    break
                header.append(line)
            else:
                raise IOError(
                    "header line with 'track type=bedGraph' not found."
                )

            # Create the instance with autochromosomes
            array = cls(
                "auto-write",
                stranded=strand != ".",
                typecode=typecode,
                storage='step',
                header=''.join(header),
            )

            # Load contents
            for line in f:
                chrom, start, end, value = line.rstrip('\n\r').split()
                start, end = int(start), int(end)
                if typecode in ('i', 'l'):
                    value = int(value)
                elif typecode == 'd':
                    value = float(value)
                else:
                    raise ValueError(f"Typecode not supported: {typecode}")

                iv = GenomicInterval(chrom, start, end, strand=strand)
                array[iv] = value

        finally:
            # Close the file only if we were the ones to open it
            if not hasattr(file_or_filename, "read"):
                f.close()

        return array

    def write_bigwig_file(
            self,
            filename,
            strand='.',
            ):
        '''Write GenomicArray to BigWig file

        BigWig files are used to visualize genomic "tracks", notably in
        UCSC's genomic viewer. They are, in a sense, the binary compressed
        equivalent of BedGraph files. This function stores the GenomicArray
        into such a file for further use.

        Args:
            filename (str or path): Where to store the data.
            strand (".", "+", or "-"): Which strand to write to file.

        The BigWig file format is described here:

            http://genome.ucsc.edu/goldenPath/help/bigWig.html

        NOTE: This function requires the package pyBigWig at:

            https://github.com/deeptools/pyBigWig

        Install it via pip, conda, or see instructions at that page.
        '''
        try:
            import pyBigWig
        except ImportError:
            raise ImportError(
                'pyBigWig is required to write a GenomicArray to a bigWig file',
            )

        if (not self.stranded) and strand != ".":
            raise ValueError("Strand specified in unstranded GenomicArray.")
        if self.stranded and strand not in (strand_plus, strand_minus):
            raise ValueError("Strand must be specified for stranded GenomicArray.")

        with pyBigWig.open(filename, "w") as bw:
            # Write header with chromosome info
            header = []
            for chrom in self.chrom_vectors:
                cv = self.chrom_vectors[chrom][strand]
                end = cv.iv.end
                header.append((chrom, end))
            bw.addHeader(header)

            # Write data (use a buffer for efficiency)
            entries = {'chrom': [], 'start': [], 'ends': [], 'values': []}
            bufsize = 1000

            def write_with_buffer(bw, entries, bufsize, newentry=None):
                if newentry is not None:
                    for key in entries:
                        entries[key].append(newentry[key])
                if len(entries) >= bufsize:
                    bw.addEntries(
                        entries['chrom'], entries['start'],
                        ends=entries['ends'], values=entries['values'],
                    )
                    for key in entries:
                        entries[key].clear()

            for chrom in self.chrom_vectors:
                cv = self.chrom_vectors[chrom][strand]
                for iv, value in cv.steps():
                    if iv.start == -sys.maxsize - 1 or iv.end == sys.maxsize:
                        continue

                    entry = {
                        'chrom': chrom,
                        'start': iv.start,
                        'ends': iv.end,
                        'values': value,
                    }
                    write_with_buffer(bw, entries, bufsize, newentry=entry)
            # Flush buffer
            write_with_buffer(bw, entries, bufsize=1)

    @classmethod
    def from_bigwig_file(cls, filename, strand=".", typecode="d"):
        '''Create GenomicArray from BigWig file

        See GenomicArray.write_bigwig for details on the file format.

        Args:
            filename (str or path): Where to load the data from.
            strand ("+", "-", or "."): strandedness of the returned array.
            typecode ("d", "i", or "l"): Type of data in the file.
              "d" means floating point (double), "i" is integer, "l" is long
              integer.

        Returns:
            A GenomicArray instance with the data.
        '''
        try:
            import pyBigWig
        except ImportError:
            raise ImportError(
                'pyBigWig is required to write a GenomicArray to a bigWig file',
            )

        # Avoid circular import
        from HTSeq import BigWig_Reader

        with BigWig_Reader(filename) as bw:
            chrom_dict = dict(bw.chroms())

            # Create the instance with specified chromosomes
            array = cls(
                chrom_dict,
                stranded=strand != ".",
                typecode=typecode,
                storage='step',
            )

            for chrom in chrom_dict:
                intervals = bw.intervals(chrom, raw=True)
                for i, (start, end, value) in enumerate(intervals):
                    # Set the chromosome offset with the first value, since
                    # they are ordered. The StepVector will be offset compared
                    # to that.
                    if i == 0:
                        array.chrom_vectors[chrom][strand].offset = start
                        array.chrom_vectors[chrom][strand].iv.start = start

                    # Bypass GenomicArray.__setitem__ for efficiency
                    # This can be done because, unlike for BedGraph files,
                    # we know the length of chromosomes a priori from the
                    # header.
                    array.chrom_vectors[chrom][strand][start: end] = value

        return array

    def steps(self):
        '''Get the steps, independent of storage method

        Each "step" is a GenomicInterval with fixed value of the array. For
        instance, if we have 3 counts on chromosome '1' between 0 and 10
        (exclded) and 6 counts between 10 and 20 (end of chromosome), we would
        get two steps: (0, 10, 3) and (10, 20, 6). If the GenomicArray is
        stranded, genomic intervals of the appropriate strandedness are
        returned.
        '''
        return _HTSeq_internal.GenomicArray_steps(self)


def _GenomicArray_unpickle(stranded, typecode, chrom_vectors):
    ga = GenomicArray({}, stranded, typecode)
    ga.chrom_vectors = chrom_vectors
    return ga


###########################
# Sequences
###########################


def _make_translation_table_for_complementation():
    return bytes.maketrans(b'ACGTacgt', b'TGCAtgca')


cdef bytes _translation_table_for_complementation = _make_translation_table_for_complementation()


cpdef bytes reverse_complement(bytes seq):
    """Returns the reverse complement of DNA sequence 'seq'. Does not yet
    work with extended IUPAC nucleotide letters or RNA."""

    return seq[::-1].translate(_translation_table_for_complementation)


base_to_column = {'A': 0, 'C': 1, 'G': 2, 'T': 3, 'N': 4}


cdef class Sequence(object):
    """A Sequence, typically of DNA, with a name."""

    def __init__(self, bytes seq, str name="unnamed"):
        self.seq = seq
        self.name = name
        self.descr = None

    cpdef Sequence get_reverse_complement(self, bint rename=True):
        if rename:
            return Sequence(
                reverse_complement(self.seq),
                "revcomp_of_" + self.name)
        else:
            return Sequence(
                reverse_complement(self.seq),
                self.name)

    def __str__(self):
        return self.seq.decode()

    def __repr__(self):
        return "<%s object '%s' (length %d)>" % (
            self.__class__.__name__, self.name, len(self.seq))

    def __len__(self):
        return len(self.seq)

    def __getitem__(self, item):
        if self.name.endswith("[part]"):
            new_name = self.name
        else:
            new_name = self.name + "[part]"
        return Sequence(self.seq[item], new_name)

    def __getstate__(self):
        return {'seq': self.seq,
                'name': self.name,
                'descr': self.descr}

    def __setstate__(self, state):
        self.seq = state['seq']
        self.name = state['name']
        self.descr = state['descr']

    def __reduce__(self):
        return (self.__class__, (self.seq, self.name), self.__getstate__())

    def write_to_fasta_file(self, fasta_file, characters_per_line=70):
        """Write sequence to file in FASTA format.

        Arguments:
                - fasta_file (file handle): destination file
                    - characters_per_line (int >=0): if 0, write the whole sequence on a single line. Otherwise, break into several lines if the sequence is long enough.

        """
        if self.descr is not None:
            fasta_file.write(">%s %s\n" % (self.name, self.descr))
        else:
            fasta_file.write(">%s\n" % self.name)

        if characters_per_line == 0:
            fasta_file.write(self.seq.decode() + "\n")
        else:
            i = 0
            while i * characters_per_line < len(self.seq):
                fasta_file.write(
                    self.seq[i * characters_per_line: (i + 1) * characters_per_line].decode() + "\n")
                i += 1

    cpdef object add_bases_to_count_array(Sequence self, numpy.ndarray count_array_):

        cdef numpy.ndarray[numpy.int64_t, ndim = 2] count_array = count_array_
        cdef int seq_length = len(self.seq)

        if numpy.PyArray_DIMS(count_array)[0] < seq_length:
            raise ValueError("'count_array' too small for sequence.")
        if numpy.PyArray_DIMS(count_array)[1] < 5:
            raise ValueError("'count_array' has too few columns.")

        cdef numpy.npy_intp i
        cdef char b
        cdef char * seq_cstr = self.seq
        for i in range(seq_length):
            b = seq_cstr[i]
            if b in [b'A', b'a']:
                count_array[i, 0] += 1
            elif b in [b'C', b'c']:
                count_array[i, 1] += 1
            elif b in [b'G', b'g']:
                count_array[i, 2] += 1
            elif b in [b'T', b't']:
                count_array[i, 3] += 1
            elif b in [b'N', b'n', b'.']:
                count_array[i, 4] += 1
            else:
                raise ValueError("Illegal base letter encountered.")

        return None

    cpdef Sequence trim_left_end(Sequence self, Sequence pattern, float mismatch_prop=0.):
        cdef int seqlen = len(self.seq)
        cdef int patlen = len(pattern.seq)
        cdef int minlen
        if seqlen < patlen:
            minlen = seqlen
        else:
            minlen = patlen
        cdef char * seq_cstr = self.seq
        cdef char * pat_cstr = pattern.seq
        cdef int match = 0
        cdef int i, j
        cdef int num_mismatches
        for i in range(1, minlen + 1):
            num_mismatches = 0
            for j in range(i):
                if seq_cstr[j] != pat_cstr[patlen - i + j]:
                    num_mismatches += 1
                    if num_mismatches > mismatch_prop * i:
                        break
            else:
                match = i
        return self[match: seqlen]

    cpdef Sequence trim_right_end(Sequence self, Sequence pattern, float mismatch_prop=0.):
        cdef int seqlen = len(self.seq)
        cdef int patlen = len(pattern.seq)
        cdef int minlen
        if seqlen < patlen:
            minlen = seqlen
        else:
            minlen = patlen
        cdef char * seq_cstr = self.seq
        cdef char * pat_cstr = pattern.seq
        cdef int match = 0
        cdef int i, j
        cdef int num_mismatches
        for i in range(1, minlen + 1):
            num_mismatches = 0
            for j in range(i):
                if seq_cstr[seqlen - i + j] != pat_cstr[j]:
                    num_mismatches += 1
                    if num_mismatches > mismatch_prop * i:
                        break
            else:
                match = i
        return self[0: seqlen - match]


cdef class SequenceWithQualities(Sequence):
    """A Sequence with base-call quality scores.
    It now has property  'qual', an integer NumPy array of Sanger/Phred
    quality scores of the  base calls.
    """

    def __init__(self, bytes seq, str name, bytes qualstr, str qualscale="phred"):
        """ Construct a SequenceWithQuality object.

          seq       - The actual sequence.
          name      - The sequence name or ID
          qualstr   - The quality string. Must have the same length as seq
          qualscale - The encoding scale of the quality string. Must be one of
                        "phred", "solexa", "solexa-old", or "noquals" )
        """
        Sequence.__init__(self, seq, name)
        if qualscale != "noquals":
            if len(seq) != len(qualstr):
                raise ValueError("'seq' and 'qualstr' do not have the same length.")
            self._qualstr = qualstr
        else:
            self._qualstr = b''
        self._qualscale = qualscale
        self._qualarr = None
        self._qualstr_phred = b''

    cdef _fill_qual_arr(SequenceWithQualities self):
        cdef int seq_len = len(self.seq)
        if self._qualscale == "missing":
            raise ValueError("Quality string missing.")
        if seq_len != len(self._qualstr):
            raise ValueError("Quality string has not the same length as sequence.")
        cdef numpy.ndarray[numpy.uint8_t, ndim= 1] qualarr = numpy.empty((seq_len, ), numpy.uint8)
        cdef int i
        cdef char * qualstr = self._qualstr
        if self._qualscale == "phred":
            for i in range(seq_len):
                qualarr[i] = qualstr[i] - 33
        elif self._qualscale == "solexa":
            for i in range(seq_len):
                qualarr[i] = qualstr[i] - 64
        elif self._qualscale == "solexa-old":
            for i in range(seq_len):
                qualarr[i] = 10 * \
                    math.log10(1 + 10 ** (qualstr[i] - 64) / 10.0)
        else:
            raise ValueError("Illegal quality scale '%s'." % self._qualscale)
        self._qualarr = qualarr

    property qual:
        def __get__(self):
            if self._qualarr is None:
                self._fill_qual_arr()
            return self._qualarr

        def __set__(self, newvalue):
            if not (isinstance(newvalue, numpy.ndarray) and newvalue.dtype == numpy.uint8):
                raise TypeError("qual can only be assigned a numpy array of type numpy.uint8")
            if not (newvalue.shape == (len(self.seq), )):
                raise TypeError("assignment to qual with illegal shape")
            self._qualarr = newvalue
            self._qualstr = b""
            self._qualscale = "none"
            self._qualstr_phred = b""
            # Experimentally trying to set qualstr when the array is modified
            # directly
            tmp = self.qualstr
            self._qualstr = self._qualstr_phred

    def __repr__(self):
        return "<%s object '%s'>" % (self.__class__.__name__, self.name)

    def __getitem__(self, item):
        if self.name.endswith("[part]"):
            new_name = self.name
        else:
            new_name = self.name + "[part]"
        return SequenceWithQualities(
            self.seq[item], new_name, self.qualstr[item])

    @property
    def qualstr(self):
        cdef int seqlen
        cdef char * qualstr_phred_cstr = self._qualstr_phred
        cdef int i
        cdef numpy.ndarray[numpy.uint8_t, ndim = 1] qual_array
        if qualstr_phred_cstr[0] == 0:
            if self._qualscale == "noquals":
                raise ValueError("Quality string missing")
            if self._qualscale == "phred":
                self._qualstr_phred = self._qualstr
            else:
                seqlen = len(self.seq)
                # FIXME: is this fixed now?
                self._qualstr_phred = b' ' * seqlen
                qualstr_phred_cstr = self._qualstr_phred
                if self._qualarr is None:
                    self._fill_qual_arr()
                qual_array = self._qualarr
                for i in range(seqlen):
                    qualstr_phred_cstr[i] = 33 + qual_array[i]
        return self._qualstr_phred

    def write_to_fastq_file(self, fastq_file):
        if hasattr(self, "descr") and self.descr is not None:
            fastq_file.write("@%s %s\n" % (self.name, self.descr))
        else:
            fastq_file.write("@%s\n" % self.name)
        fastq_file.write(self.seq.decode() + "\n")
        fastq_file.write("+\n")
        fastq_file.write(self.qualstr.decode() + "\n")

    def get_fastq_str(self, bint convert_to_phred=False):
        sio = cStringIO.StringIO()
        self.write_to_fastq_file(sio, convert_to_phred)
        return sio.getvalue()

    cpdef SequenceWithQualities get_reverse_complement(self, bint rename=True):
        cdef SequenceWithQualities res
        if rename:
            res = SequenceWithQualities(
                reverse_complement(self.seq),
                "revcomp_of_" + self.name,
                self._qualstr[::-1],
                self._qualscale)
        else:
            res = SequenceWithQualities(
                reverse_complement(self.seq),
                self.name,
                self._qualstr[::-1],
                self._qualscale)
        if self._qualarr is not None:
            res._qualarr = self._qualarr[::-1]
        return res

    cpdef object add_qual_to_count_array(SequenceWithQualities self,
                                         numpy.ndarray count_array_):

        cdef numpy.ndarray[numpy.int64_t, ndim = 2] count_array = count_array_
        if self._qualarr is None:
            self._fill_qual_arr()
        cdef numpy.ndarray[numpy.uint8_t, ndim = 1] qual_array = self._qualarr

        cdef numpy.npy_intp seq_length = numpy.PyArray_DIMS(qual_array)[0]
        cdef numpy.npy_intp qual_size = numpy.PyArray_DIMS(count_array)[1]

        if seq_length > numpy.PyArray_DIMS(count_array)[0]:
            raise ValueError("'count_array' too small for sequence.")

        cdef numpy.npy_intp i
        cdef numpy.npy_int q
        for i in range(seq_length):
            q = qual_array[i]
            if(q >= qual_size):
                raise ValueError("Too large quality value encountered.")
            count_array[i, q] += 1

        return None

    cpdef SequenceWithQualities trim_left_end_with_quals(SequenceWithQualities self,
                                                         Sequence pattern, int max_mm_qual=5):
        cdef int seqlen = len(self.seq)
        cdef int patlen = len(pattern.seq)
        cdef int minlen
        if seqlen < patlen:
            minlen = seqlen
        else:
            minlen = patlen
        cdef char * seq_cstr = self.seq
        cdef char * pat_cstr = pattern.seq
        cdef int match = 0
        cdef int i, j
        cdef int sum_mm_qual
        if self._qualarr is None:
            self._fill_qual_arr()
        cdef numpy.ndarray[numpy.uint8_t, ndim = 1] qual_array = self._qualarr
        for i in range(1, minlen + 1):
            num_mismatches = 0
            for j in range(i):
                if seq_cstr[j] != pat_cstr[patlen - i + j]:
                    sum_mm_qual += qual_array[j]
                    if sum_mm_qual > max_mm_qual:
                        break
            else:
                match = i
        return self[match: seqlen]

    cpdef SequenceWithQualities trim_right_end_with_quals(SequenceWithQualities self,
                                                          Sequence pattern, int max_mm_qual=5):
        cdef int seqlen = len(self.seq)
        cdef int patlen = len(pattern.seq)
        cdef int minlen
        if seqlen < patlen:
            minlen = seqlen
        else:
            minlen = patlen
        cdef char * seq_cstr = self.seq
        cdef char * pat_cstr = pattern.seq
        cdef int match = 0
        cdef int i, j
        cdef int sum_mm_qual
        if self._qualarr is None:
            self._fill_qual_arr()
        cdef numpy.ndarray[numpy.uint8_t, ndim = 1] qual_array = self._qualarr
        for i in range(1, minlen + 1):
            sum_mm_qual = 0
            for j in range(i):
                if seq_cstr[seqlen - i + j] != pat_cstr[j]:
                    sum_mm_qual += qual_array[seqlen - i + j]
                    if sum_mm_qual > max_mm_qual:
                        break
            else:
                match = i
        return self[0: seqlen - match]


###########################
# CIGAR codes (SAM format)
###########################
_re_cigar_codes = re.compile('([MIDNSHP=X])')

cigar_operation_names = {
    'M': 'matched',
    'I': 'inserted',
    'D': 'deleted',
    'N': 'skipped',
    'S': 'soft-clipped',
    'H': 'hard-clipped',
    'P': 'padded',
    '=': 'sequence-matched',
    'X': 'sequence-mismatched'}


cigar_operation_codes = ['M', 'I', 'D', 'N', 'S', 'H', 'P', '=', 'X']
cigar_operation_code_dict = dict(
    [(x, i) for i, x in enumerate(cigar_operation_codes)])


cdef class CigarOperation(object):

    cdef public str type
    cdef public int size
    cdef public GenomicInterval ref_iv
    cdef public int query_from, query_to

    def __init__(self, str type_, int size, int rfrom, int rto, int qfrom,
                 int qto, str chrom, str strand, bint check=True):
        self.type = type_
        self.size = size
        self.ref_iv = GenomicInterval(chrom, rfrom, rto, strand)
        self.query_from = qfrom
        self.query_to = qto
        if check and not self.check():
            raise ValueError("Inconsistent CIGAR operation.")

    def __repr__(self):
        return "< %s: %d base(s) %s on ref iv %s, query iv [%d,%d) >" % (
            self.__class__.__name__, self.size, cigar_operation_names[
                self.type],
            str(self.ref_iv), self.query_from, self.query_to)

    def check(CigarOperation self):
        cdef int qlen = self.query_to - self.query_from
        cdef int rlen = self.ref_iv.length
        if self.type == 'M' or self.type == '=' or self.type == 'X':
            if not (qlen == self.size and rlen == self.size):
                return False
        elif self.type == 'I' or self.type == 'S':
            if not (qlen == self.size and rlen == 0):
                return False
        elif self.type == 'D' or self.type == 'N':
            if not (qlen == 0 and rlen == self.size):
                return False
        elif self.type == 'H' or self.type == 'P':
            if not (qlen == 0 and rlen == 0):
                return False
        else:
            return False
        return True


cpdef list parse_cigar(str cigar_string, int ref_left=0, str chrom="", str strand="."):
    cdef list split_cigar, cl
    cdef int size
    cdef str code
    split_cigar = _re_cigar_codes.split(cigar_string)
    if split_cigar[-1] != '' or len(split_cigar) % 2 != 1:
        raise ValueError("Illegal CIGAR string '%s'" % cigar_string)
    cl = []
    for i in range(len(split_cigar) // 2):
        try:
            size = int(split_cigar[2 * i])
        except ValueError:
            raise ValueError("Illegal CIGAR string '%s'" % cigar_string)
        code = split_cigar[2 * i + 1]
        cl.append((code, size))
    return build_cigar_list(cl, ref_left, chrom, strand)


cpdef list build_cigar_list(list cigar_pairs, int ref_left=0, str chrom="", str strand="."):
    cdef list split_cigar, res
    cdef int rpos, qpos, size
    cdef str code
    rpos = ref_left
    qpos = 0
    res = []
    for code, size in cigar_pairs:
        if code == 'M' or code == '=' or code == 'X':
            res.append(CigarOperation(
                code, size, rpos, rpos + size, qpos, qpos + size, chrom, strand))
            rpos += size
            qpos += size
        elif code == 'I':
            res.append(CigarOperation(
                'I', size, rpos, rpos, qpos, qpos + size, chrom, strand))
            qpos += size
        elif code == 'D':
            res.append(CigarOperation(
                'D', size, rpos, rpos + size, qpos, qpos, chrom, strand))
            rpos += size
        elif code == 'N':
            res.append(CigarOperation(
                'N', size, rpos, rpos + size, qpos, qpos, chrom, strand))
            rpos += size
        elif code == 'S':
            res.append(CigarOperation(
                'S', size, rpos, rpos, qpos, qpos + size, chrom, strand))
            qpos += size
        elif code == 'H':
            res.append(CigarOperation(
                'H', size, rpos, rpos, qpos, qpos, chrom, strand))
        elif code == 'P':
            res.append(CigarOperation(
                'P', size, rpos, rpos, qpos, qpos, chrom, strand))
        else:
            raise ValueError("Unknown CIGAR code '%s' encountered." % code)
    return res


###########################
# Alignment
###########################
cdef class Alignment(object):
    """An aligned read (typically from a BAM file).

    An alignment object can be defined in different ways but will always
    provide these attributes:
      read:      a SequenceWithQualities object with the read
      aligned:   whether the read is aligned
      iv:        a GenomicInterval object with the alignment position
    """

    def __init__(self, read, iv):
        self._read = read
        self.iv = iv

    @property
    def read(self):
        return self._read

    def __repr__(self):
        cdef str s
        if self.paired_end:
            s = "Paired-end read"
        else:
            s = "Read"
        if self.aligned:
            return "<%s object: %s '%s' aligned to %s>" % (
                self.__class__.__name__, s, self.read.name, str(self.iv))
        else:
            return "<%s object: %s '%s', not aligned>" % (
                self.__class__.__name__, s, self.read.name)

    @property
    def paired_end(self):
        return False

    @property
    def aligned(self):
        """Returns True unless self.iv is None. The latter indicates that
        this record decribes a read for which no alignment was found.
        """
        return self.iv is not None


cdef class AlignmentWithSequenceReversal(Alignment):
    """Many aligners report the read's sequence in reverse-complemented form
    when it was mapped to the reverse strand. For such alignments, a
    daughter class of this one should be used.

    Then, the read is stored as aligned in the 'read_as_aligned' field,
    and get reverse-complemented back to the sequenced form when the 'read'
    attribute is sequenced.
    """

    def __init__(self, SequenceWithQualities read_as_aligned, GenomicInterval iv):
        self.read_as_aligned = read_as_aligned
        self._read_as_sequenced = None
        self.iv = iv

    property read:
        def __get__(self):
            if self._read_as_sequenced is None:
                if (not self.aligned) or self.iv.strand != "-":
                    self._read_as_sequenced = self.read_as_aligned
                else:
                    self._read_as_sequenced = self.read_as_aligned.get_reverse_complement()
                    self._read_as_sequenced.name = self.read_as_aligned.name
            return self._read_as_sequenced
        # def __set__( self, read ):
        #   self.read_as_aligned = read
        #   self._read_as_sequenced = None


cdef class BowtieAlignment(AlignmentWithSequenceReversal):
    """When reading in a Bowtie file, objects of the class BowtieAlignment
    are returned. In addition to the 'read' and 'iv' fields (see Alignment
    class), the fields 'reserved' and 'substitutions' are provided. These
    contain the content of the respective columns of the Bowtie output

    [A parser for the substitutions field will be added soon.]
    """

    cdef public str reserved
    cdef public str substitutions

    def __init__(self, bowtie_line):
        cdef str readId, strand, chrom, position, read, qual
        cdef int positionint
        (readId, strand, chrom, position, read, qual,
         self.reserved, self.substitutions) = bowtie_line.split('\t')
        positionint = int(position)
        AlignmentWithSequenceReversal.__init__(self,
                                               SequenceWithQualities(
                                                   read, readId, qual),
                                               GenomicInterval(chrom, positionint, positionint + len(read), strand))


cdef _parse_SAM_optional_field_value(str field):
    if len(field) < 5 or field[2] != ':' or field[4] != ':':
        raise ValueError("Malformatted SAM optional field '%'" % field)
    if field[3] == 'A':
        return field[5]
    elif field[3] == 'i':
        return int(field[5:])
    elif field[3] == 'f':
        return float(field[5:])
    elif field[3] == 'Z':
        return field[5:]
    elif field[3] == 'H':
        return int(field[5:], 16)
    elif field[3] == 'B':
        if field[5] == 'f':
            return numpy.array(field[7:].split(','), float)
        else:
            return numpy.array(field[7:].split(','), int)
    else:
        raise ValueError("SAM optional field with illegal type letter '%s'" % field[2])


cdef class SAM_Alignment(AlignmentWithSequenceReversal):
    """When reading in a SAM file, objects of the class SAM_Alignment
    are returned. In addition to the 'read', 'iv' and 'aligned' fields (see
    Alignment class), the following fields are provided:
     - aQual: the alignment quality score
     - cigar: a list of CigarOperatio objects, describing the alignment
     - tags: the extra information tags [not yet implemented]
    """

    def to_pysam_AlignedSegment(self, sf):
        try:
            import pysam
        except ImportError:
            sys.stderr.write(
                "Please Install PySam to use this functionality (http://code.google.com/p/pysam/)")
            raise

        a = pysam.AlignedSegment()
        a.query_sequence = self.read.seq if self.iv == None or self.iv.strand == '+' else self.read.get_reverse_complement().seq
        a.query_qualities = self.read.qual if self.iv == None or self.iv.strand == '+' else self.read.get_reverse_complement().qual
        a.query_name = self.read.name
        a.flag = self.flag
        a.tags = self.optional_fields
        if self.aligned:
            a.cigartuples = [(cigar_operation_code_dict[c.type], c.size)
                             for c in self.cigar]
            a.reference_start = self.iv.start
            a.reference_id = sf.gettid(self.iv.chrom)
            a.template_length = self.inferred_insert_size
            a.mapping_quality = self.aQual
        else:
            a.reference_start = -1
            a.reference_id = -1
        if self.mate_aligned:
            a.next_reference_id = sf.gettid(self.mate_start.chrom)
            a.next_reference_start = self.mate_start.start
        else:
            a.next_reference_id = -1
            a.next_reference_start = -1
        return a

    def to_pysam_AlignedRead(self, sf):
        try:
            import pysam
        except ImportError:
            sys.stderr.write(
                "Please Install PySam to use this functionality (http://code.google.com/p/pysam/)")
            raise

        a = pysam.AlignedRead()
        a.seq = self.read.seq
        a.qual = self.read.qualstr
        a.qname = self.read.name
        a.flag = self.flag
        a.tags = self.optional_fields
        if self.aligned:
            a.cigar = [(cigar_operation_code_dict[c.type], c.size)
                       for c in self.cigar]
            a.pos = self.iv.start
            a.tid = sf.gettid(self.iv.chrom)
            a.isize = self.inferred_insert_size
            a.mapq = self.aQual
        else:
            a.pos = -1
            a.tid = -1
        if self.mate_aligned:
            a.mrnm = sf.gettid(self.mate_start.chrom)
            a.mpos = self.mate_start.start
        else:
            a.mrnm = -1
            a.mpos = -1
        return a

    @classmethod
    def from_pysam_AlignedRead(cls, read, samfile):
        strand = "-" if read.is_reverse else "+"
        if not read.is_unmapped:
            chrom = samfile.getrname(read.tid)
            iv = GenomicInterval(chrom, read.pos, read.aend, strand)
        else:
            iv = None
        if (read.qual is None) or (read.qual == "*"):
            seq = SequenceWithQualities(
                read.query_sequence.encode(), read.qname, b'',
                "noquals")
        else:
            seq = SequenceWithQualities(
                read.query_sequence.encode(), read.qname, read.qual.encode(),
                )

        a = SAM_Alignment(seq, iv)
        a.cigar = build_cigar_list([(cigar_operation_codes[code], length) for (
            code, length) in read.cigar], read.pos, chrom, strand) if iv != None else []
        a.inferred_insert_size = read.isize
        a.aQual = read.mapq
        a.flag = read.flag
        a.proper_pair = read.is_proper_pair
        a.not_primary_alignment = read.is_secondary
        a.failed_platform_qc = read.is_qcfail
        a.pcr_or_optical_duplicate = read.is_duplicate
        a.supplementary = read.is_supplementary
        a.original_sam_line = ""
        a.optional_fields = read.tags
        if read.is_paired:
            # These two should be but are not always consistent
            if (not read.mate_is_unmapped) and (read.mrnm != -1):
                strand = "-" if read.mate_is_reverse else "+"
                a.mate_start = GenomicPosition(
                    samfile.getrname(read.mrnm), read.mpos, strand)
            else:
                a.mate_start = None
            if read.is_read1:
                a.pe_which = intern("first")
            elif read.is_read2:
                a.pe_which = intern("second")
            else:
                a.pe_which = intern("unknown")
        else:
            a.pe_which = intern("not_paired_end")
        return a

    @classmethod
    def from_pysam_AlignedSegment(cls, read, samfile):
        strand = "-" if read.is_reverse else "+"
        if not read.is_unmapped:
            chrom = samfile.getrname(read.tid)
            iv = GenomicInterval(chrom, read.reference_start,
                                 read.reference_end, strand)
        else:
            iv = None

        # read.query_sequence can be empty (e.g. nanopore runs), then pysam
        # casts it as None or *. We recast it as an empty string to preserve
        #types. It is then converted to ASCII
        query_sequence = read.query_sequence
        if (query_sequence is None) or (query_sequence == '*'):
            query_sequence = ''
        query_sequence = query_sequence.encode()

        # read.qual can be empty (e.g. special filtering, artificial), then
        # it comes through as None or as a * (see previous comment). In this
        # case, we cast it as an empty ASCII _and_ we have to tell the
        # class about the issue.
        if (read.qual is None) or (read.qual == "*"):
            seq = SequenceWithQualities(
                query_sequence, read.query_name, b'',
                'noquals')
        else:
            seq = SequenceWithQualities(
                query_sequence, read.qname, read.qual.encode(),
                )

        a = SAM_Alignment(seq, iv)
        a.cigar = build_cigar_list([(cigar_operation_codes[code], length) for (
            code, length) in read.cigartuples], read.reference_start, chrom, strand) if iv != None else []
        a.inferred_insert_size = read.template_length
        a.aQual = read.mapping_quality
        a.flag = read.flag
        a.proper_pair = read.is_proper_pair
        a.not_primary_alignment = read.is_secondary
        a.failed_platform_qc = read.is_qcfail
        a.pcr_or_optical_duplicate = read.is_duplicate
        a.supplementary = read.is_supplementary
        a.original_sam_line = ""
        a.optional_fields = read.tags
        if read.is_paired:
            # These two should be but are not always consistent
            if (not read.mate_is_unmapped) and (read.mrnm != -1):
                strand = "-" if read.mate_is_reverse else "+"
                a.mate_start = GenomicPosition(samfile.getrname(
                    read.mrnm), read.next_reference_start, strand)
            else:
                a.mate_start = None
            if read.is_read1:
                a.pe_which = intern("first")
            elif read.is_read2:
                a.pe_which = intern("second")
            else:
                a.pe_which = intern("unknown")
        else:
            a.pe_which = intern("not_paired_end")
        return a

    @classmethod
    def from_SAM_line(cls, line):
        cdef str qname, flag, rname, pos, mapq, cigar,
        cdef str mrnm, mpos, isize, seq, qual
        cdef list optional_fields
        cdef int posint, flagint
        cdef str strand
        cdef list cigarlist
        cdef SequenceWithQualities swq

        fields = line.rstrip().split("\t")
        if len(fields) < 10:
            raise ValueError("SAM line does not contain at least 11 tab-delimited fields.")
        (qname, flag, rname, pos, mapq, cigar, mrnm, mpos, isize,
         seq, qual) = fields[0:11]
        optional_fields = fields[11:]

        if seq.count("=") > 0:
            raise ValueError("Sequence in SAM file contains '=', which is not supported.")
        if seq.count(".") > 0:
            raise ValueError("Sequence in SAM file contains '.', which is not supported.")
        flagint = int(flag)

        if flagint & 0x0004:     # flag "query sequence is unmapped"
            iv = None
            cigarlist = None
        else:
            if rname == "*":
                raise ValueError("Malformed SAM line: RNAME == '*' although flag bit &0x0004 cleared")
            # SAM is one-based, but HTSeq is zero-based!
            posint = int(pos) - 1
            if flagint & 0x0010:      # flag "strand of the query"
                strand = "-"
            else:
                strand = "+"
            cigarlist = parse_cigar(cigar, posint, rname, strand)
            iv = GenomicInterval(
                rname, posint, cigarlist[-1].ref_iv.end, strand)

        if qual != "*":
            swq = SequenceWithQualities(
                seq.upper().encode(), qname, qual.upper().encode())
        else:
            swq = SequenceWithQualities(
                seq.upper().encode(), qname, b"", "noquals")

        alnmt = SAM_Alignment(swq, iv)
        alnmt.flag = flagint
        alnmt.cigar = cigarlist
        alnmt.optional_fields = [
            (field[:2], _parse_SAM_optional_field_value(field)) for field in optional_fields]
        alnmt.aQual = int(mapq)
        alnmt.inferred_insert_size = int(isize)
        alnmt.original_sam_line = line

        if flagint & 0x0001:         # flag "read is paired in sequencing"
            if flagint & 0x0008:      # flag "mate is unmapped"
                alnmt.mate_start = None
            else:
                if mrnm == "*":
                    raise ValueError("Malformed SAM line: MRNM == '*' although flag bit &0x0008 cleared")
                posint = int(mpos) - 1
                if flagint & 0x0020:   # flag "strand of the mate"
                    strand = "-"
                else:
                    strand = "+"
                alnmt.mate_start = GenomicPosition(mrnm, posint, strand)
                if alnmt.mate_start.chrom == "=":
                    if alnmt.iv is not None:
                        alnmt.mate_start.chrom = alnmt.iv.chrom
            if flagint & 0x0040:
                alnmt.pe_which = intern("first")
            elif flagint & 0x0080:
                alnmt.pe_which = intern("second")
            else:
                alnmt.pe_which = intern("unknown")
        else:
            alnmt.mate_start = None
            alnmt.pe_which = intern("not_paired_end")

        alnmt.proper_pair = flagint & 0x0002 > 0
        alnmt.not_primary_alignment = flagint & 0x0100 > 0
        alnmt.failed_platform_qc = flagint & 0x0200 > 0
        alnmt.pcr_or_optical_duplicate = flagint & 0x0400 > 0
        alnmt.supplementary = flagint & 0x0800 > 0

        return alnmt

    property flag:
        def __get__(self):
            return self._flag

        def __set__(self, value):
            self._flag = value

    @property
    def paired_end(self):
        return self.pe_which != "not_paired_end"

    @property
    def mate_aligned(self):
        return self.mate_start is not None

    def get_sam_line(self):
        cdef str cigar = ""
        cdef GenomicInterval query_start, mate_start
        cdef CigarOperation cop

        if self.aligned:
            query_start = self.iv
        else:
            query_start = GenomicPosition("*", -1)

        if self.mate_start is not None:
            mate_start = self.mate_start
        else:
            mate_start = GenomicPosition("*", -1)

        if self.cigar is not None:
            for cop in self.cigar:
                cigar += str(cop.size) + cop.type
        else:
            cigar = "*"

        return '\t'.join(
                (self.read.name,
                 str(self.flag),
                 query_start.chrom,
                 str(query_start.start + 1),
                 str(self.aQual),
                 cigar,
                 mate_start.chrom,
                 str(mate_start.pos + 1),
                 str(self.inferred_insert_size),
                 self.read_as_aligned.seq.decode(),
                 self.read_as_aligned.qualstr.decode(),
                 '\t'.join(self.raw_optional_fields())))

    def has_optional_field(SAM_Alignment self, str tag):
        '''Check if this alignment has the specified optional field

        Args:
            tag: the field to look for.
        Returns: a bool with True if the field has been found, False otherwise.
        '''
        for p in self.optional_fields:
            if p[0] == tag:
                return True
        return False

    def optional_field(SAM_Alignment self, str tag):
        res = [p for p in self.optional_fields if p[0] == tag]
        if len(res) == 1:
            return res[0][1]
        else:
            if len(res) == 0:
                raise KeyError("SAM optional field tag %s not found" % tag)
            else:
                raise ValueError("SAM optional field tag %s not unique" % tag)

    def raw_optional_fields(self):
        res = []
        for op in self.optional_fields:
            if op[1].__class__ == str:
                if len(op[1]) == 1:
                    tc = "A"
                else:
                    tc = "Z"
            elif op[1].__class__ == int:
                tc = "i"
            elif op[1].__class__ == float:
                tc = "f"
            else:
                tc = "H"
            res.append(":".join([op[0], tc, str(op[1])]))
        return res


###########################
# Helpers
###########################
cpdef list quotesafe_split(bytes s, bytes split=b';', bytes quote=b'"'):
    cdef list l = []
    cdef int i = 0
    cdef int begin_token = 0
    cdef bint in_quote = False
    cdef char * s_c = s
    cdef char split_c = split[0]
    cdef char quote_c = quote[0]
    if len(split) != 1:
        raise ValueError("'split' must be length 1")
    if len(quote) != 1:
        raise ValueError("'quote' must be length 1")
    while s_c[i] != 0:
        if s_c[i] == quote_c:
            in_quote = not in_quote
        elif (not in_quote) and s_c[i] == split_c:
            l.append(s[begin_token: i])
            begin_token = i + 1
        i += 1
    l.append(s[begin_token:])
    if in_quote:
        raise ValueError("unmatched quote")
    return l



from collections import namedtuple
import numpy as np
import warnings


Interval = namedtuple('Interval', ['start', 'end'])


class StretchVector:
    """Sparse representation for 'island' of dense data on a long line.

    This class is the most basic object representing 'windows of data' in a sea
    of unknowns of a linear structure, e.g. a chromosome. The basic design is to
    store a list of (start, end) intervals, each of them associated with a
    "stretch", i.e. a dense numpy array.

    In its simplest form, a StretchVector can be thought of as an array with
    an arbitratily large offset. This can be useful e.g. to plot coverage or
    signal enrichments while keeping real genomic coordinates.

    The class does more heavy lifting, however, when several separate arrays
    are used, e.g. to represent "peaks" of signals (ATAC-Seq, ChIP-Seq, etc.).
    In that situation, the bookkeeping involved in manipulating all offsets
    correctly can become burdensome quickly, so StretchVector takes care of it.
    In addition, extension of data to flanks is easy, so if you want to expand
    an existing data window by 1kb on each side, that can be done directly:

    >>> sv = HTSeq.StretchVector(typecode="d")
    >>> sv[6789: 8900] = 56.8
    >>> sv[6000: 7000] = 20 # <-- left extension
    >>> sv[8000: 9000] = 10 # <-- right extension

    You can use StretchVector as a storage option for higher level objects such
    as ChromVector and GenomicArray. Those classes support strandedness, unlike
    StretchVector itself.
    """
    _typecode_dict = {
        "d": np.float32,
        "i": np.int32,
        "l": np.int64,
        "O": object,
    }

    def __init__(self, typecode):
        """Create an empty StretchVector of a given type

        Args:
            typecode ("d", "i", "l", or "O"): The dtype of the stored data. Can
              be "d" (double, i.e. np.float32), "i" (np.int32), "l" (np.int64),
              or "O" (generic object). Note that np.float64 data will be recast
              to np.float32, which might lead to a loss of machine precision.

        Returns:
            A StretchVector instance of the chosen type.
        """
        self.typecode = typecode
        self.ivs = []
        self.stretches = []

    def _in_stretch(self, index):
        if len(self.stretches) == 0:
            return -1

        if isinstance(index, int):
            if (index < self.ivs[0].start) or (index >= self.ivs[-1].end):
                return -1
            for i, iv in enumerate(self.ivs):
                if index < iv.start:
                    return -1
                if index < iv.end:
                    return i

    def _get_interval(self, start, end):
        if len(self.stretches) == 0:
            return self

        ivs = []
        stretches = []
        for i, iv in enumerate(self.ivs):
            # They end before the start, skip
            if iv.end <= start:
                continue

            # They start after the end, skip all remaining
            if iv.start >= end:
                break

            # This interval overlap with start-end
            if (iv.start <= start) and (iv.end <= end):
                new_iv = Interval(start, iv.end)
                new_stretch = self.stretches[i][start - iv.start:]
                ivs.append(new_iv)
                stretches.append(new_stretch)
                continue

            if (iv.start <= start) and (iv.end > end):
                new_iv = Interval(start, end)
                new_stretch = self.stretches[i][start - iv.start:-(iv.end - end)]
                ivs.append(new_iv)
                stretches.append(new_stretch)
                break

            if (iv.start > start) and (iv.end <= end):
                new_iv = Interval(iv.start, iv.end)
                new_stretch = self.stretches[i]
                ivs.append(new_iv)
                stretches.append(new_stretch)
                continue

            if (iv.start > start) and (iv.end > end):
                new_iv = Interval(iv.start, end)
                new_stretch = self.stretches[i][:-(iv.end - end)]
                ivs.append(new_iv)
                stretches.append(new_stretch)
                break

        new_cls = self.__class__(self.typecode)
        new_cls.ivs = ivs
        new_cls.stretches = stretches
        return new_cls

    def _set_interval(self, start, end, values):
        if len(self.stretches) == 0:
            self.ivs.append(Interval(start, end))
            self.stretches.append(
                np.zeros(end - start, self._typecode_dict[self.typecode])
            )
            self.stretches[-1][:] = values
            return len(self.ivs) - 1

        # For each end, there are two possibilities, inside or outside an
        # existing stretch
        idx_start = self._in_stretch(start)
        idx_end = self._in_stretch(end - 1)

        # Neither is in, make a new stretch and delete existing stretches
        if idx_start == idx_end == -1:
            new_iv = Interval(start, end)
            new_stretch = np.zeros(end - start, self._typecode_dict[self.typecode])
            new_stretch[:] = values
            new_ivs = []
            new_stretches = []
            new_added = False
            for i, iv in enumerate(self.ivs):
                # Stretches before
                if start >= iv.end:
                    new_ivs.append(iv)
                    new_stretches.append(self.stretches[i])
                    continue

                # Add new stretch
                if not new_added:
                    new_ivs.append(new_iv)
                    new_stretches.append(new_stretch)

                # Skip overlapping stretches
                if (start <= iv.start) and (end >= iv.end):
                    continue

                # Stretches after
                new_ivs.append(iv)
                new_stretches.append(self.stretches[i])

            # Add new stretch if still missing
            if not new_added:
                new_ivs.append(new_iv)
                new_stretches.append(new_stretch)

        # Start is in a stretch, end is not
        elif (idx_start != -1) and (idx_end == -1):
            new_iv = Interval(self.ivs[idx_start].start, end)
            new_stretch = np.zeros(
                end - self.ivs[idx_start].start,
                self._typecode_dict[self.typecode],
            )
            l1 = start - self.ivs[idx_start].start
            new_stretch[:l1] = self.stretches[idx_start][:l1]
            new_stretch[l1:] = values

            new_ivs = self.ivs[:idx_start] + [new_iv]
            new_stretches = self.stretches[:idx_start] = [new_stretch]

            for i, iv in enumerate(self.ivs[idx_start:], idx_start):
                # Skip the first one
                if i == idx_start:
                    continue

                # Skip overlapping stretches
                if iv.end < end:
                    continue

                # Stretches after
                new_ivs.append(iv)
                new_stretches.append(self.stretches[i])

        # Start is not in a stretch, end is
        elif (idx_start == -1) and (idx_end != -1):
            new_iv = Interval(start, self.ivs[idx_end].end)
            new_stretch = np.zeros(
                self.ivs[idx_end].end - start,
                self._typecode_dict[self.typecode],
            )
            l2 = self.ivs[idx_end].end - end
            if l2 == 0:
                l2 = -len(self.stretches[idx_end])
            new_stretch[:-l2] = values
            if -l2 != len(self.stretches[idx_end]):
                new_stretch[-l2:] = self.stretches[idx_end][-l2:]

            new_ivs = []
            new_stretches = []
            for i, iv in enumerate(self.ivs):
                # Stretches before
                if start >= iv.end:
                    new_ivs.append(iv)
                    new_stretches.append(self.stretches[i])
                    continue
                break

            # New stretch
            new_ivs.append(new_iv)
            new_stretches.append(new_stretch)

            # If there are stretches left, add them
            if idx_end != len(self.ivs) - 1:
                new_ivs.extend(self.ivs[idx_end+1:])
                new_stretches.extend(self.stretches[idx_end+1:])

        # Both start and end are in the same stretch
        elif idx_start == idx_end:
            l1 = start - self.ivs[idx_start].start
            l2 = self.ivs[idx_end].end - end
            if l2 == 0:
                l2 = -len(self.stretches[idx_end])
            new_ivs = self.ivs
            new_stretches = self.stretches
            new_stretches[idx_start][l1:-l2] = values

        # They are in different stretches
        else:
            new_iv = Interval(
                    self.ivs[idx_start].start,
                    self.ivs[idx_end].end,
                )
            new_stretch = np.zeros(
                self.ivs[idx_end].end - self.ivs[idx_start].start,
                self._typecode_dict[self.typecode],
            )
            l1 = start - self.ivs[idx_start].start
            l2 = self.ivs[idx_end].end - end
            if l2 == 0:
                l2 = -len(self.stretches[idx_end])
            new_stretch[:l1] = self.stretches[idx_start][:l1]
            new_stretch[l1:-l2] = values
            if -l2 != len(self.stretches[idx_end]):
                new_stretch[-l2:] = self.stretches[idx_end][-l2:]

            new_ivs = self.ivs[:idx_start] + [iv]
            if idx_end != len(self.ivs) - 1:
                new_ivs.extend(self.ivs[idx_end+1:])
                new_stretches.extend(self.stretches[idx_end+1:])

        self.ivs = new_ivs
        self.stretches = new_stretches

    def _add_stretch(self, start, end, i_start=0):
        for i, iv in enumerate(self.ivs, i_start):
            if start < iv.start:
                self.ivs.insert(
                    i,
                    Interval(start, end),
                )
                self.stretches.insert(
                    i,
                    np.zeros(end - start, self._typecode_dict[self.typecode])
                )
                return i

        self.ivs.append(
            Interval(start, end),
        )
        self.stretches.append(
            np.zeros(end - start, self._typecode_dict[self.typecode])
        )
        return len(self.ivs) - 1

    def __getitem__(self, index):
        """Get a view of a portion of the StretchVector


        Args:
            index (int, slice, or GenomicInterval): Coordinate or interval to
              extract. For slices, the stretches from this intervals are
              *views* of the original array, so changes in them will be
              reflected in the parent StretchVector as well.
        Returns:
            A number of index is an int, containing the value at that site.
            A StretchVector with adapted stretches if index is anything else.

        """
        from HTSeq import GenomicInterval

        if isinstance(index, int):
            idx_iv = self._in_stretch(index)
            if idx_iv == -1:
                return None
            return self.stretches[idx_iv][index - self.ivs[idx_iv].start]

        elif isinstance(index, slice):
            if index.step is not None and index.step != 1:
                raise ValueError(
                        "Striding slices (i.e., step != 1) are not supported")
            if index.start is None:
                index.start = 0
            if index.stop is None:
                if len(self.ivs) == 0:
                    raise IndexError('No stretches, cannot find end')
                index.stop = self.ivs[-1].end

            return self._get_interval(index.start, index.stop)

        elif isinstance(index, GenomicInterval):
            return self.__getitem__(slice(index.start, index.end))

    def __setitem__(self, index, values):
        """Set/reset values within or outside the stretch

        Args:
            index (int, slice, or GenomicInterval): Coordinate or interval to
              set. These can be within current stretches, between them, outside,
              or partially overlapping.
            values (numpy.ndarray or convertible sequence): Values to set or
              reset at those locations. This will be recast as a numpy array
              of the appropriate dtype within this function.

        Returns: None
        """
        from HTSeq import GenomicInterval

        # Leave dtype out for now for speed, it will be taken care of later on
        values = np.asarray(values)

        if isinstance(index, int):
            idx_iv = self._in_stretch(index)
            if idx_iv == -1:
                idx_iv = self._add_stretch(index, index + 1)
            self.stretches[idx_iv][index - self.ivs[idx_iv].start] = values
            return

        elif isinstance(index, slice):
            if index.step is not None and index.step != 1:
                raise ValueError(
                        "Striding slices (i.e., step != 1) are not supported")
            if index.start is None:
                index.start = 0
            if index.stop is None:
                if len(self.ivs) == 0:
                    raise IndexError('No stretches, cannot find end')
                index.stop = self.ivs[-1].end

            self._set_interval(index.start, index.stop, values)

        elif isinstance(index, GenomicInterval):
            return self.__setitem__(slice(index.start, index.end), values)


    def todense(self):
        """Dense numpy array of the whole stretch, using NaNs for missing data"""
        if len(self.ivs) == 0:
            return np.empty(0, self._typecode_dict[self.typecode])

        if len(self.ivs) == 1:
            return self.stretches[0].copy()

        # At least two stretches, have to stitch them
        start = self.ivs[0].start
        res = np.empty(
            self.ivs[-1].end - start,
            self.stretches[0].dtype,
        )
        res[:] = np.nan

        for i, iv in enumerate(self.ivs):
            res[iv.start - start: iv.end - start] = self.stretches[i]

        return res

    @classmethod
    def from_dense(cls, array, offset=0):
        """Create from dense array with NaNs

        Args:
            array (numpy.array): dense array containing NaNs at positions to
              be skipped.
            offset (int): Start of the initial interval.
        """
        if array.dtype == np.float64:
            warnings.warn("np.float64 array converted to np.float32")
            array = array.astype(np.float32)

        for typecode, dtype in cls._typecode_dict.items():
            if dtype == array.dtype:
                sv = cls(typecode=typecode)
                break
        else:
            raise TypeError('Typecode not found for dtype: '+str(array.dtype))

        if len(array) == 0:
            return sv

        flips = np.diff(np.isnan(array)).nonzero()[0]

        # No flips: either all good or all skip
        if flips.sum() == 0:
            if np.isnan(array[0]):
                return sv
            sv.ivs.append(Interval(offset, offset + len(array)))
            sv.stretches.append(array.copy())
            return sv

        # If we start with nan, just increase the offset
        if np.isnan(array[0]):
            add_offset = flips[0] + 1
            offset += add_offset
            array = array[add_offset:]

            # Single flip means keep the whole rest
            if len(flips) > 1:
                flips = flips[1:]
                flips -= add_offset
            else:
                sv.ivs.append(Interval(offset, offset + len(array)))
                sv.stretches.append(array.copy())
                return sv

        # Now we have at least one flip left, and we start with a number/object
        # If we have an odd number of flips, we can forget the last block
        if len(flips) % 2:
            end = flips[-1] + 1
            array = array[:end]
            flips = flips[:-1]

        # No flip left, all good
        if len(flips) == 0:
            sv.ivs.append(Interval(offset, offset + len(array)))
            sv.stretches.append(array.copy())
            return sv

        # Now we start and end with a number/object, and there are at least
        # two flips. Initial stretch
        sv.ivs.append(Interval(offset, offset + flips[0] + 1))
        sv.stretches.append(
                array[:flips[0] + 1]
        )
        # Intermediate stretches
        for i in range((len(flips) // 2) - 1):
            new_iv = Interval(offset + flips[i * 2 + 1] + 1, offset + flips[i * 2 + 2] + 1)
            new_stretch = array[flips[i * 2 + 1] + 1: flips[i * 2 + 2] + 1]
            sv.ivs.append(new_iv)
            sv.stretches.append(new_stretch)
        # Final stretch
        new_iv = Interval(offset + flips[-1] + 1, offset + len(array))
        new_stretch = array[flips[-1] + 1:]
        sv.ivs.append(new_iv)
        sv.stretches.append(new_stretch)

        return sv

    def __iter__(self):
        """Iterate over intervals and stretches ("islands")."""
        return zip(self.ivs, self.stretches)

    def copy(self):
        """Make a copy the StretchVector and of all its stretches"""
        sv_new = StretchVector(typecode=self.typecode)
        for iv, stretch in self:
            sv_new.ivs.append(Interval(iv.start, iv.end))
            sv_new.stretches.append(stretch.copy())
        return sv_new

    def shift(self, offset):
        """Shift all stretch intervals by a constant number

        Args:
            offset (int): Shift the start and end coordinates of each stretch
              by this amount.
        Returns:
            None. The function acts in place.
        """
        for i, iv in self.ivs:
            self.ivs[i] = Interval(iv.start + offset, iv.end + offset)



import sys
import numpy as np


class UnknownChrom(Exception):
    pass


def my_showwarning(message, category, filename, lineno=None, file=None,
                   line=None):
    sys.stderr.write("Warning: %s\n" % message)


def invert_strand(iv):
    iv2 = iv.copy()
    if iv2.strand == "+":
        iv2.strand = "-"
    elif iv2.strand == "-":
        iv2.strand = "+"
    else:
        raise ValueError("Illegal strand")
    return iv2


def _merge_counts(
        results,
        attributes,
        additional_attributes,
        sparse=False,
        dtype=np.float32,
        ):
    barcodes = 'cell_barcodes' in results

    if barcodes:
        cbs = results['cell_barcodes']
        counts = results['counts']

    feature_attr = sorted(attributes.keys())
    other_features = [
        ('__no_feature', 'empty'),
        ('__ambiguous', 'ambiguous'),
        ('__too_low_aQual', 'lowqual'),
        ('__not_aligned', 'notaligned'),
        ('__alignment_not_unique', 'nonunique'),
        ]

    fea_names = [fea for fea in feature_attr] + [fea[0] for fea in other_features]
    L = len(fea_names)
    if barcodes:
        n = len(cbs)
    else:
        n = len(results)
    if not sparse:
        table = np.zeros(
            (n, L),
            dtype=dtype,
        )
    else:
        from scipy.sparse import lil_matrix
        table = lil_matrix((n, L), dtype=dtype)

    if not barcodes:
        fea_ids = [fea for fea in feature_attr] + [fea[1] for fea in other_features]
        for j, r in enumerate(results):
            for i, fn in enumerate(fea_ids):
                if i < len(feature_attr):
                    countji = r['counts'][fn]
                else:
                    countji = r[fn]
                if countji > 0:
                    table[j, i] = countji
    else:
        for j, cb in enumerate(cbs):
            for i, fn in enumerate(fea_names):
                countji = counts[cb][fn]
                if countji > 0:
                    table[j, i] = countji

    if sparse:
        table = table.tocsr()

    feature_metadata = {
        'id': fea_names,
    }
    for iadd, attr in enumerate(additional_attributes):
        feature_metadata[attr] = [attributes[fn][iadd] for fn in feature_attr]

    return {
        'feature_metadata': feature_metadata,
        'table': table,
    }


def _count_results_to_tsv(
        results,
        samples_name,
        attributes,
        additional_attributes,
        output_filename,
        output_delimiter,
        output_append=False,
        add_tsv_header=False
        ):

    barcodes = 'cell_barcodes' in results

    pad = ['' for attr in additional_attributes]

    if barcodes:
        cbs = results['cell_barcodes']
        counts = results['counts']

        # Print or write header
        fields = [''] + pad + cbs
        line = output_delimiter.join(fields)
        if output_filename == '':
            print(line)
        else:
            with open(output_filename, 'w') as f:
                f.write(line)
                f.write('\n')

    elif add_tsv_header:
        # Write the header.
        # Only get here if we don't have cell barcodes, i.e. this is not called by htseq-count-barcode,
        # and user wants the tsv header
        file_header = output_delimiter.join([''] + pad + samples_name)

        if output_filename == '':
            print(file_header)
        else:
            # If append to existing file, then open as a
            file_open_opt = 'a' if output_append else 'w'

            with open(output_filename, file_open_opt) as f:
                f.write(file_header)
                f.write('\n')

    # Each feature is a row with feature id, additional attrs, and counts
    feature_attr = sorted(attributes.keys())
    for ifn, fn in enumerate(feature_attr):
        if not barcodes:
            fields = [fn] + attributes[fn] + [str(r['counts'][fn]) for r in results]
        else:
            fields = [fn] + attributes[fn] + [str(counts[cb][fn]) for cb in cbs]

        line = output_delimiter.join(fields)
        if output_filename == '':
            print(line)
        else:
            omode = 'a' if output_append or (ifn > 0) or barcodes or add_tsv_header else 'w'
            with open(output_filename, omode) as f:
                f.write(line)
                f.write('\n')

    # Add other features (unmapped, etc.)
    other_features = [
        ('__no_feature', 'empty'),
        ('__ambiguous', 'ambiguous'),
        ('__too_low_aQual', 'lowqual'),
        ('__not_aligned', 'notaligned'),
        ('__alignment_not_unique', 'nonunique'),
        ]
    for title, fn in other_features:
        if not barcodes:
            fields = [title] + pad + [str(r[fn]) for r in results]
        else:
            fields = [title] + pad + [str(counts[cb][title]) for cb in cbs]
        line = output_delimiter.join(fields)
        if output_filename == '':
            print(line)
        else:
            with open(output_filename, 'a') as f:
                f.write(line)
                f.write('\n')


def _count_table_to_mtx(
        filename,
        table,
        feature_metadata,
        samples,
        ):
    if not str(filename).endswith('.mtx'):
        raise ValueError('Matrix Marker filename should end with ".mtx"')

    try:
        from scipy.io import mmwrite
    except ImportError:
        raise ImportError('Install scipy for mtx support')

    filename_pfx = str(filename)[:-4]
    filename_feature_meta = filename_pfx+'_features.tsv'
    filename_samples = filename_pfx+'_samples.tsv'

    # Write main matrix (features as columns)
    mmwrite(
        filename,
        table,
    )

    # Write input filenames
    with open(filename_samples, 'wt') as fout:
        for fn in samples:
            fout.write(fn+'\n')

    # Write feature metadata (ids and additional attributes)
    with open(filename_feature_meta, 'wt') as fout:
        nkeys = len(feature_metadata)
        for ik, key in enumerate(feature_metadata):
            if ik != nkeys - 1:
                fout.write(key+'\t')
            else:
                fout.write(key+'\n')
        nfeatures = len(feature_metadata[key])
        for i in range(nfeatures):
            for ik, key in enumerate(feature_metadata):
                if ik != nkeys - 1:
                    fout.write(feature_metadata[key][i]+'\t')
                else:
                    fout.write(feature_metadata[key][i]+'\n')


def _count_table_to_h5ad(
        filename,
        table,
        feature_metadata,
        samples,
        ):
    try:
        import anndata
    except ImportError:
        raise ImportError('Install the anndata package for h5ad support')

    # If they have anndata, they have scipy and pandas too
    import pandas as pd

    # We don't have additional attribute (e.g. gene name) for htseq specific features like __no_feature.
    # Hence the trick is to convert the array to series so the value for htseq specific features like __no_feature
    # column is set NaN.
    # See: https://stackoverflow.com/questions/19736080/creating-dataframe-from-a-dictionary-where-entries-have-different-lengths
    feature_metadata = pd.DataFrame(dict([(k, pd.Series(v)) for k, v in feature_metadata.items()]))
    feature_metadata.set_index(feature_metadata.columns[0], inplace=True)

    adata = anndata.AnnData(
        X=table,
        obs=pd.DataFrame([], index=samples),
        var=feature_metadata,
    )
    adata.write_h5ad(filename)


def _count_table_to_loom(
        filename,
        table,
        feature_metadata,
        samples,
        ):

    try:
        import loompy
    except ImportError:
        raise ImportError('Install the loompy package for loom support')

    # Loom uses features as rows...
    layers = {'': table.T}
    row_attrs = feature_metadata
    col_attrs = {'_index': samples}
    loompy.create(
        filename,
        layers=layers,
        row_attrs=row_attrs,
        col_attrs=col_attrs,
    )


def _write_output(
    results,
    samples,
    attributes,
    additional_attributes,
    output_filename,
    output_delimiter,
    output_append,
    sparse=False,
    dtype=np.float32,
    add_tsv_header=False
    ):

    """
    Export the gene counts as tsv/csv, mtx, loom, h5ad files.
    Note, need to update the parameter documentations.

    Parameters
    ----------
    results : list
        List of dictionaries with each element representing the counts for an input BAM file.
        Note, the list is in order of the samples parameter. So the first element in the list corresponds to
        the first file in samples parameter.
    samples : list
        List of input BAM files.

    """

    # Write output to stdout or TSV/CSV
    if output_filename == '':
        _count_results_to_tsv(
            results,
            samples,
            attributes,
            additional_attributes,
            output_filename,
            output_delimiter,
            output_append=False,
            add_tsv_header=add_tsv_header
        )
        return

    # Get file extension/format
    output_sfx = output_filename.split('.')[-1].lower()

    if output_sfx in ('csv', 'tsv'):
        _count_results_to_tsv(
            results,
            samples,
            attributes,
            additional_attributes,
            output_filename,
            output_delimiter,
            output_append,
            add_tsv_header=add_tsv_header
        )
        return

    # Make unified object of counts and feature metadata
    output_dict = _merge_counts(
        results,
        attributes,
        additional_attributes,
        sparse=sparse,
        dtype=dtype,
    )

    if output_sfx == 'mtx':
        _count_table_to_mtx(
            output_filename,
            output_dict['table'],
            output_dict['feature_metadata'],
            samples,
        )
        return

    if output_sfx == 'loom':
        _count_table_to_loom(
            output_filename,
            output_dict['table'],
            output_dict['feature_metadata'],
            samples,
        )
        return

    if output_sfx == 'h5ad':
        _count_table_to_h5ad(
            output_filename,
            output_dict['table'],
            output_dict['feature_metadata'],
            samples,
        )
        return

    raise ValueError(
        f'Format not recognized for output count file: {output_sfx}')



import os
import subprocess as sp
import unittest
import numpy as np
import pysam
import pytest
import conftest
import pandas as pd
import math

try:
    import scipy
except ImportError:
    scipy = None

try:
    import anndata
except ImportError:
    anndata = None

try:
    import loompy
except ImportError:
    loompy = None


data_folder = conftest.get_data_folder()


def load_result_file(filename):
    sfx = filename.split('.')[-1]
    if sfx in ('csv', 'tsv'):
        with open(filename, 'r') as f:
            result = f.read()
    elif sfx == 'mtx':
        from scipy.io import mmread
        result = mmread(filename)
    elif sfx == 'h5ad':
        result = anndata.read_h5ad(filename)
    elif sfx == 'loom':
        result = loompy.connect(filename)
    else:
        raise ValueError(f'File extension not supported: {sfx}')

    return {'result': result, 'fmt': sfx}


def close_file(filename, resultd):
    fmt = resultd['fmt']
    if fmt == 'loom':
        resultd['result'].close()


class HTSeqCountBase(unittest.TestCase):
    def _customAssertEqual(self, outputd, expectedd):
        output_fmt = outputd['fmt']
        expected_fmt = expectedd['fmt']
        self.assertEqual(output_fmt, expected_fmt)

        fmt = output_fmt
        output = outputd['result']
        expected = expectedd['result']

        if fmt in ('tsv', 'csv'):
            self.assertEqual(output, expected)
        elif fmt == 'mtx':
            self.assertIsNone(
                np.testing.assert_array_equal(
                    output, expected,
                ))
        elif fmt == 'loom':
            self.assertIsNone(
                np.testing.assert_array_equal(
                    output[:, :], expected[:, :],
                ))
            #TODO: metadata and filenames
        elif fmt == 'h5ad':
            self.assertIsNone(
                np.testing.assert_array_equal(
                    output.X, output.X,
                ))
            #TODO: metadata and filenames
        else:
            raise ValueError(f'Format not supported: {fmt}')

    def _run(self, t, remove_res_files=True):
        expected_fn = t.get('expected_fn', None)
        expected_stderr = t.get('expected_stderr', None)
        call = t['call']

        # Replace with injected variable
        call = [x.replace(f'{data_folder}/', data_folder) for x in call]
        if expected_fn is not None:
            expected_fn = expected_fn.replace(f'{data_folder}/', data_folder)

        ## local testing
        #if call[0] == 'htseq-count':
        #    call = ['python', 'HTSeq/scripts/count.py'] + call[1:]
        #else:
        #    call = ['python', 'HTSeq/scripts/count_with_barcodes.py'] + call[1:]

        print(' '.join(call))
        p = sp.run(
            ' '.join(call),
            shell=True,
            check=True,
            capture_output=True,
        )
        if expected_stderr:
            actual_stderr_str = p.stderr.decode()
            self.assertEqual(actual_stderr_str, expected_stderr)
        output = p.stdout.decode()

        if '-c' in call:
            output_fn = call[call.index('-c') + 1]
            output = load_result_file(output_fn)
        else:
            output = {'result': output, 'fmt': 'tsv'}
            output_fn = None

        if expected_fn is None:
            if '--version' in call:
                print('version:', output['result'])
            return

        if not os.path.isfile(expected_fn):
            print('Missing output file, creating one in current folder')
            out_fn = os.path.basename(expected_fn)
            if output_fn is None:
                with open(out_fn, 'wt') as f:
                    f.write(output['result'])
            else:
                import shutil
                shutil.copy(output_fn, out_fn)
            pytest.fail(
                'Expected filename not found, output filename copied in {out_fn}',
            )

        expected = load_result_file(expected_fn)

        try:
            self._customAssertEqual(output, expected)
        finally:
            if output_fn is not None:
                close_file(output_fn, output)
                close_file(expected_fn, expected)
                # FIXME
                if remove_res_files:#output['fmt'] not in ['h5ad', 'loom']:
                    os.remove(output_fn)


class HTSeqCount(HTSeqCountBase):
    cmd = 'htseq-count'

    def test_version(self):
        self._run({
            'call': [
                self.cmd,
                '--version'],
            })

    def test_simple(self):
        self._run({
            'call': [
                self.cmd,
                f'{data_folder}/bamfile_no_qualities.sam',
                f'{data_folder}/bamfile_no_qualities.gtf',
            ],
            'expected_fn': f'{data_folder}/bamfile_no_qualities.tsv',
            })

    def test_output_tsv(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                f'{data_folder}/bamfile_no_qualities.sam',
                f'{data_folder}/bamfile_no_qualities.gtf',
                ],
            'expected_fn': f'{data_folder}/bamfile_no_qualities.tsv',
            })

    @unittest.skipIf(scipy is None, "test case depends on scipy")
    def test_output_mtx(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.mtx',
                f'{data_folder}/bamfile_no_qualities.sam',
                f'{data_folder}/bamfile_no_qualities.gtf',
                ],
            'expected_fn': f'{data_folder}/bamfile_no_qualities.mtx',
            })

    @unittest.skipIf(anndata is None, "test case depends on anndata")
    def test_output_h5ad(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.h5ad',
                f'{data_folder}/bamfile_no_qualities.sam',
                f'{data_folder}/bamfile_no_qualities.gtf',
                ],
            'expected_fn': f'{data_folder}/bamfile_no_qualities.h5ad',
            })

    @unittest.skipIf(loompy is None, "test case depends on loompy")
    def test_output_loom(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.loom',
                f'{data_folder}/bamfile_no_qualities.sam',
                f'{data_folder}/bamfile_no_qualities.gtf',
                ],
            'expected_fn': f'{data_folder}/bamfile_no_qualities.loom',
            })

    def test_output_tsv_header(self):
        # Header must be there
        self._run({
            'call': [
                self.cmd,
                '--with-header', ' ',
                '-c', f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
                f'{data_folder}/10x_pbmc1k/3_cells/cell1.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell2.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell3.bam',
                f'{data_folder}/10x_pbmc1k/HomoSapiens.GRCh38-2020-A_subsampled.gtf',
            ],
            'expected_fn': f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
        }, remove_res_files=False)

        try:
            dat = pd.read_csv(f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv', delimiter='\t')
            self.assertEqual(len(dat.columns), 4)
            self.assertTrue(f'{data_folder}10x_pbmc1k/3_cells/cell1.bam' in dat.columns)
            self.assertTrue(f'{data_folder}10x_pbmc1k/3_cells/cell2.bam' in dat.columns)
            self.assertTrue(f'{data_folder}10x_pbmc1k/3_cells/cell3.bam' in dat.columns)
        finally:
            os.remove(f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv')

    def test_output_tsv_no_header(self):
        # Backwards compatibility of no header
        self._run({
            'call': [
                self.cmd,
                '-c', f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
                f'{data_folder}/10x_pbmc1k/3_cells/cell1.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell2.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell2.bam',
                f'{data_folder}/10x_pbmc1k/HomoSapiens.GRCh38-2020-A_subsampled.gtf',
            ],
            'expected_fn': f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
        }, remove_res_files=False)
        try:
            dat = pd.read_csv(f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv', delimiter='\t', header=None)
            self.assertEqual(dat.shape[0], 7)
            self.assertEqual(dat.loc[0,0], 'ENSG00000188976')
        finally:
            os.remove(f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv')

    def test_output_tsv_header_and_append(self):
        # Make sure append still works properly
        self._run({
            'call': [
                self.cmd,
                '--with-header', ' ',
                '-c', f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
                f'{data_folder}/10x_pbmc1k/3_cells/cell1.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell2.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell3.bam',
                f'{data_folder}/10x_pbmc1k/HomoSapiens.GRCh38-2020-A_subsampled.gtf',
            ],
            'expected_fn': f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
        }, remove_res_files=False)

        self._run({
            'call': [
                self.cmd,
                '--append-output', ' ',
                '-c', f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
                f'{data_folder}/10x_pbmc1k/3_cells/cell1.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell2.bam',
                f'{data_folder}/10x_pbmc1k/3_cells/cell3.bam',
                f'{data_folder}/10x_pbmc1k/HomoSapiens.GRCh38-2020-A_subsampled.gtf',
            ],
            'expected_fn': f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv',
        }, remove_res_files=False)
        try:
            dat = pd.read_csv(f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv', delimiter='\t')
            self.assertEqual(dat.shape[0], 14)
            self.assertTrue(f'{data_folder}10x_pbmc1k/3_cells/cell1.bam' in dat.columns)
            self.assertTrue(f'{data_folder}10x_pbmc1k/3_cells/cell2.bam' in dat.columns)
            self.assertTrue(f'{data_folder}10x_pbmc1k/3_cells/cell3.bam' in dat.columns)
        finally:
            os.remove(f'{data_folder}/10x_pbmc1k/test_tsv_header.tsv')


    # Testing multiple cores on travis makes a mess
    #{'call': [
    #    'htseq-count',
    #    '-n', '2',
    #    f'{data_folder}/bamfile_no_qualities.sam',
    #    f'{data_folder}/bamfile_no_qualities.gtf',
    #    ],
    # 'expected_fn': f'{data_folder}/bamfile_no_qualities.tsv'},

    def test_no_qualities(self):
        self._run({
            'call': [
                self.cmd,
                f'{data_folder}/bamfile_no_qualities.bam',
                f'{data_folder}/bamfile_no_qualities.gtf',
            ],
            'expected_fn': f'{data_folder}/bamfile_no_qualities.tsv',
            })

    def test_some_missing_sequences(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                f'{data_folder}/yeast_RNASeq_excerpt_some_empty_seqs.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_some_empty_seqs.tsv',
            })

    def test_intersection_nonempty(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts.tsv',
            })

    def test_feature_query(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                '--feature-query', '\'gene_id == "YPR036W-A"\'',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_YPR036W-A.tsv',
            })

    def test_additional_attributes(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                '--additional-attr', 'gene_name',
                '--additional-attr', 'exon_number',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_additional_attributes.tsv',
            })

    def test_multiple_and_additional_attributes(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '-i', 'gene_id',
                '-i', 'exon_number',
                '--additional-attr', 'gene_name',
                '--additional-attr', 'exon_number',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_exon_level_and_additional_attributes.tsv',
            })

    def test_additional_attributes_chromosome_info(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                '--additional-attr', 'gene_name',
                '--additional-attr', 'exon_number',
                '--add-chromosome-info',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_additional_attributes_chromosome_info.tsv',
            })

    @unittest.skipIf(anndata is None, "test case depends on anndata")
    def test_additional_attributes_h5ad(self):
        # Get gene name as additional attr but output as h5ad file.
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                '--additional-attr', 'gene_name',
                '--additional-attr', 'exon_number',
                '-c', f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes_counts.h5ad',
                f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes.sam',
                f'{data_folder}/10x_pbmc1k/HomoSapiens.GRCh38-2020-A_subsampled.gtf',
                ],
            'expected_fn': f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes_counts.h5ad',
            }, remove_res_files=False)

        # Check gene name and exon number is there
        try:
            dat = anndata.read_h5ad(f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes_counts.h5ad')
            self.assertEqual(dat.var.loc['ENSG00000188976', 'gene_name'], 'NOC2L')
            self.assertEqual(dat.var.loc['ENSG00000251562', 'gene_name'], 'MALAT1')
            self.assertEqual(dat.var.loc['ENSG00000188976', 'exon_number'], '2')
            self.assertEqual(dat.var.loc['ENSG00000251562', 'exon_number'], '2')

            htseq_specific_features = ['__no_feature', '__ambiguous', '__too_low_aQual',
                                       '__not_aligned', '__alignment_not_unique']
            for feat in htseq_specific_features:
                self.assertTrue(math.isnan(dat.var.loc[feat, 'gene_name']))
                self.assertTrue(math.isnan(dat.var.loc[feat, 'exon_number']))

        finally:
            # clean up
            os.remove(data_folder + "/10x_pbmc1k/subsampled_with_missing_barcodes_counts.h5ad")

    def test_nonunique_fraction(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'fraction',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_nonunique_fraction.tsv',
            })

    def test_withNH(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'all',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_nonunique.tsv',
            })

    def test_twocolumns(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '-i', 'gene_id',
                '--additional-attr', 'gene_name',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_twocolumns.tsv',
            })

    def test_ignore_secondary(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'ignore',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withNH.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withNH_counts_ignore_secondary.tsv',
            })

    def test_no_contig_overlap_warning(self):
        """ Ensures that no contig warning is present in stderr
            @see https://github.com/htseq/htseq/issues/63 """

        # We print full file path in the error message, so have to use template & replace
        bam_file = os.path.join(data_folder, 'SRR001432_head_sorted.bam')
        stderr_template = open(f'{data_folder}/no_contig_overlap_warning_stderr.txt').read()
        expected_stderr = stderr_template.format(FILENAME=bam_file)
        self._run({
            'call': [
                self.cmd,
                bam_file,
                f'{data_folder}/bamfile_no_qualities.gtf',
                ],
                'expected_stderr': expected_stderr,
            })

    def test_contig_overlap_no_warning(self):
        """ Ensures that warning is NOT present in stderr when there are contig overlaps
            @see https://github.com/htseq/htseq/issues/63 """

        bam_file = os.path.join(data_folder, 'bamfile_no_qualities.sam')
        expected_stderr = open(f'{data_folder}/contig_overlap_no_warning_stderr.txt').read()
        self._run({
            'call': [
                self.cmd,
                bam_file,
                f'{data_folder}/bamfile_no_qualities.gtf',
                ],
                'expected_stderr': expected_stderr,
            })

    def test_default_feature_type(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                f'{data_folder}/yeast_RNASeq_excerpt.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56_multiple_types.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_default_type.tsv',
            })

    def test_non_default_feature_type(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                '--type', 'pseudogene',
                f'{data_folder}/yeast_RNASeq_excerpt.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56_multiple_types.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_non_default_type.tsv',
            })

    def test_multiple_feature_types(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                '--type', 'exon',
                '--type', 'pseudogene',
                f'{data_folder}/yeast_RNASeq_excerpt.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56_multiple_types.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_multiple_types.tsv',
            })


class HTSeqCountBarcodes(HTSeqCountBase):
    cmd = 'htseq-count-barcodes'

    def test_version(self):
        self._run({
            'call': [
                self.cmd,
                '--version'],
            })

    def test_simple(self):
        self._run({
            'call': [
                self.cmd,
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.tsv',
            })

    def test_output_tsv(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.tsv',
            })

    def test_output_tsv_chromosome_info(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                '-m', 'intersection-nonempty',
                '--add-chromosome-info',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes_chromosome_info.tsv',
            })

    def test_output_tsv_correct_UMI(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.tsv',
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                '--correct-UMI-distance', '1',
                f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes_correctUMI_1.tsv',
            })

    @unittest.skipIf(anndata is None, "test case depends on anndata")
    def test_output_h5ad(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.h5ad',
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.h5ad',
            })

    @unittest.skipIf(loompy is None, "test case depends on loompy")
    def test_output_loom(self):
        self._run({
            'call': [
                self.cmd,
                '-c', 'test_output.loom',
                '-m', 'intersection-nonempty',
                '--nonunique', 'none',
                '--secondary-alignments', 'score',
                '--supplementary-alignments', 'score',
                f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.sam',
                f'{data_folder}/Saccharomyces_cerevisiae.SGD1.01.56.gtf.gz',
                ],
            'expected_fn': f'{data_folder}/yeast_RNASeq_excerpt_withbarcodes.loom',
            })

    def test_missing_barcodes(self):
        """
        When reads are missing cell or UMI barcodes, the counts should exclude these reads
        """
        self._run({
            'call': [
                self.cmd,
                '--mode', 'union',
                '--secondary-alignments', 'ignore',
                '--supplementary-alignments', 'ignore',
                '--stranded', 'yes',
                '--counts_output_sparse',
                '-c', f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes_counts.csv',
                '--samout', f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes_counts.sam',
                f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes.sam',
                f'{data_folder}/10x_pbmc1k/HomoSapiens.GRCh38-2020-A_subsampled.gtf',
                ],
            'expected_fn': f'{data_folder}/10x_pbmc1k/subsampled_with_missing_barcodes_counts.csv',
            }, remove_res_files = False)

        # Check the result. The count csv should not register any reads with low quality
        # The SAM file should assign low quality to reads without cell or UMI barcode

        samfile = None
        try:
            df = pd.read_csv(data_folder + '/10x_pbmc1k/subsampled_with_missing_barcodes_counts.csv',
                             header=None,
                             delimiter='\t')
            df.columns = ['gene_ids', 'count']
            self.assertEqual(int(df.loc[df['gene_ids'] == 'ENSG00000188976']['count'].iloc[0]), 1)
            self.assertEqual(int(df.loc[df['gene_ids'] == 'ENSG00000251562']['count'].iloc[0]), 1)

            other_genes = ['__no_feature', '__ambiguous', '__too_low_aQual', '__not_aligned', '__alignment_not_unique']

            for g in other_genes:
                self.assertEqual(int(df.loc[df['gene_ids'] == g]['count'].iloc[0]), 0)

            samfile = pysam.AlignmentFile(data_folder + "/10x_pbmc1k/subsampled_with_missing_barcodes_counts.sam")
            read_assignments = {
                "A00228:279:HFWFVDMXX:2:1385:2085:18975": "__too_low_aQual",
                "A00228:279:HFWFVDMXX:1:2158:25464:1313": "ENSG00000251562",
                "A00228:279:HFWFVDMXX:2:1168:11153:29168": "ENSG00000188976",
                "A00228:279:HFWFVDMXX:1:1425:2781:12665": "__too_low_aQual",
                "A00228:279:HFWFVDMXX:2:2322:8621:9157": "__too_low_aQual",
                "A00228:279:HFWFVDMXX:1:2119:24270:19351": "__too_low_aQual",
                "A00228:279:HFWFVDMXX:2:2249:1045:2707": "__too_low_aQual",
                "A00228:279:HFWFVDMXX:2:2401:14172:19210": "__too_low_aQual"
            }

            for read in samfile.fetch(until_eof=True):
                qname_str = read.qname
                tag_val = read.get_tag("XF")
                self.assertEqual(tag_val, read_assignments[qname_str])

        finally:
            # clean up
            if samfile is not None:
                samfile.close()
            os.remove(data_folder + '/10x_pbmc1k/subsampled_with_missing_barcodes_counts.csv')
            os.remove(data_folder + "/10x_pbmc1k/subsampled_with_missing_barcodes_counts.sam")


if __name__ == '__main__':

    suite = HTSeqCount()
    suite.test_version()
    suite.test_simple()
    suite.test_output_tsv()
    suite.test_no_qualities()
    suite.test_intersection_nonempty()
    suite.test_feature_query()
    suite.test_additional_attributes()
    suite.test_nonunique_fraction()
    suite.test_withNH()
    suite.test_twocolumns()
    suite.test_ignore_secondary()

    suite = HTSeqCountBarcodes()
    suite.test_version()
    suite.test_missing_barcodes()



import pytest
import sys
import os
import glob
import sysconfig

build_dir = "build/lib.%s-%s" % (sysconfig.get_platform(), sys.version[0:3])
sys.path.insert(0, os.path.join(os.getcwd(), build_dir))


def test_import():
    print('Import HTSeq')
    try:
        import HTSeq
    except ModuleNotFoundError:
        print('cwd:', os.getcwd())
        print('build_dir:', build_dir)
        print('PYTHONPATH:', sys.path)


def test_version():
    print('Test version')
    import HTSeq
    print(HTSeq.__version__)
    print('Test passed')



import HTSeq
import numpy
from matplotlib import pyplot as plt


bamfile = HTSeq.BAM_Reader("SRR001432_head.bam")
gtffile = HTSeq.GFF_Reader("Homo_sapiens.GRCh37.56_chrom1.gtf")
halfwinwidth = 3000
fragmentsize = 200

tsspos = HTSeq.GenomicArrayOfSets("auto", stranded=False)
for feature in gtffile:
    if feature.type == "exon" and feature.attr["exon_number"] == "1":
        p = feature.iv.start_d_as_pos
        window = HTSeq.GenomicInterval(
            p.chrom, p.pos - halfwinwidth, p.pos + halfwinwidth, "."
        )
        tsspos[window] += p

profile = numpy.zeros(2 * halfwinwidth, dtype="i")
for almnt in bamfile:
    if almnt.aligned:
        almnt.iv.length = fragmentsize
        s = set()
        for step_iv, step_set in tsspos[almnt.iv].steps():
            s |= step_set
        for p in s:
            if p.strand == "+":
                start_in_window = almnt.iv.start - p.pos + halfwinwidth
                end_in_window = almnt.iv.end - p.pos + halfwinwidth
            else:
                start_in_window = p.pos + halfwinwidth - almnt.iv.end
                end_in_window = p.pos + halfwinwidth - almnt.iv.start
            start_in_window = max(start_in_window, 0)
            end_in_window = min(end_in_window, 2 * halfwinwidth)
            profile[start_in_window:end_in_window] += 1
